
/*
 cc -g -c imgmap.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "imgrgb.h"
#include "imgsamp.h"
#include "imgstd.h"
#include "imgexempt.h"

#include "imgmap.h"



/* ----------------------------- ImgmapcachE ------------------------------ */


int Imgmapcache_new(cache,mapname,flag)
struct ImgmapcachE **cache;
char *mapname;
int flag;
{
 *cache= TSOB_FELD(struct ImgmapcachE,1);
 if(*cache==NULL)
    return(-1);
 (*cache)->refcount= 0;
 strcpy((*cache)->mapname,mapname);
 (*cache)->start_samp= NULL;
 (*cache)->last_samp= NULL;
 return(1);
}


int Imgmapcache_destroy(del_cache,flag)
struct ImgmapcachE **del_cache;
int flag;
{
 struct ImgsamplE *samp,*next,*prev;
 struct ImgmapcachE *cache;
 int removed=0;

 cache= *del_cache;
 if(cache==NULL)
   return(0);
 cache->refcount--;
 if(cache->refcount>0)
   return(1);
 /* this should be much better while full usage of smem.c */
 for(samp= cache->last_samp;samp!=NULL;samp= prev) {
   prev= samp->prev;
   Imgsample_destroy(&samp,0);
 }

/*
 for(samp= cache->start_samp;samp!=NULL;samp= next) {
   next= samp->next;
   Imgsample_destroy(&samp,0);
 }
*/

 Smem_freE((char *) cache);
 *del_cache= NULL;
 return(2);
}


/* -------------------------------- ImgmaP ------------------------------ */


int Imgmap_new(imp,icmp,mapname,flag)
struct ImgmaP **imp;
struct ImgcmP *icmp;
char *mapname;
int flag;
{
 int ret;

 *imp= TSOB_FELD(struct ImgmaP,1);
 if(*imp==NULL)
   return(-1);
 ret= Imgmap_init(*imp,icmp,mapname,0);
 if(ret<=0)
   goto failed;
 return(1);
failed:;
 Imgmap_destroy(imp,0);
 return(-1);
}


int Imgmap_destroy(imp,flag)
struct ImgmaP **imp;
int flag;
{
 struct ImgmaP *m;
 
 m= *imp;
 if(m==NULL)
   return(0);
 Imgmap_deactivate(m,0);
 Smem_freE((char *) m);
 *imp= NULL;
 return(1);
}


int Imgmap_init(imp,icmp,mapname,flag)
struct ImgmaP *imp;
struct ImgcmP *icmp;
char *mapname;
int flag;
{
 imp->icmp= icmp;
 imp->disabled= 0;
 imp->guarding= 0;
 imp->fp_nam= imp->fp_samp= NULL;
 imp->searchpos= 0;
 imp->caching= 0;
 imp->cache= NULL;
 imp->exemptlist= NULL;
 imp->use_exemptlist= 0;
 imp->searchsamp= NULL;
 imp->spoints= Imgmap_sample_sizE;
 imp->search_method= Imgmap_search_methoD;
 imp->color_eps= Imgmap_color_epS;
 imp->bw_eps= Imgmap_bw_epS;
 imp->color_max_misses= Imgmap_color_max_misseS;
 imp->bw_max_misses= Imgmap_bw_max_misseS;
 imp->errmsg[0]= 0;
 Imgmap_set_mapname(imp,mapname,0);
 return(1);
}


int Imgmap_deactivate(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= do not close file pointer
*/
{
 if(flag&1) {
   Imgmap_fclose(imp,1);
   Imgmap_fclose(imp,2);
 }
 Imgexemptlist_destroy(&(imp->exemptlist),0);
 Imgmap_delete_cache(imp,0);
 imp->use_exemptlist= 0;
 return(1);
}


int Imgmap_copy(source,target,flag)
struct ImgmaP *source,*target;
int flag;
/*
 bit0= also copy link to cache
 bit1= target is not initialized yet
 bit2= do not copy reference to exempt list
*/
{
 if(flag&2)
   Imgmap_init(target,source->icmp,source->mapname,0);
 strcpy(target->mapname,source->mapname);
 target->disabled= source->disabled;
 target->guarding= source->guarding;
 target->fp_nam= source->fp_nam;
 target->fp_samp= source->fp_samp;
 target->searchpos= source->searchpos;
 target->spoints= source->spoints;
 target->search_method= source->search_method;
 target->color_eps= source->color_eps;
 target->bw_eps= source->bw_eps;
 target->color_max_misses= source->color_max_misses;
 target->bw_max_misses= source->bw_max_misses;
 strcpy(target->errmsg,source->errmsg);
 if(flag&1){
   target->caching= source->caching;
   Imgmap_attach_cache(target,source->cache,0);
   target->searchsamp= source->searchsamp;
 }else{
   target->caching= 0;
   target->cache= NULL;
   target->searchsamp= NULL;
 }
 if(flag&4) {
   target->exemptlist= NULL;
   target->use_exemptlist;
 } else {
   Imgexemptlist_destroy(&(target->exemptlist),0);
   Imgexemptlist_refer(source->exemptlist,&(target->exemptlist),0);
   target->use_exemptlist= source->use_exemptlist;
 }
 return(1);
}


int Imgmap_set_eps(imp,color_eps,bw_eps,flag)
struct ImgmaP *imp;
double color_eps,bw_eps;
int flag;
{
 imp->color_eps= color_eps;
 imp->bw_eps= bw_eps;
 return(1);
}


int Imgmap_set_spoints(imp,spoints,flag)
struct ImgmaP *imp;
int spoints;
int flag;
{
 imp->spoints= spoints;
 return(1);
}


int Imgmap_get_spoints(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->spoints);
}


int Imgmap_set_mapname(imp,mapname,flag)
struct ImgmaP *imp;
char *mapname;
int flag;
{
 if(strcmp(imp->mapname,mapname)==0)
   return(2);
 Imgmap_deactivate(imp,0);
 strcpy(imp->mapname,mapname);
 if(imp->caching&1) 
   Imgmap_load_cache(imp,0);
 return(1);
}


int Imgmap_get_mapname(imp,mapname,flag)
struct ImgmaP *imp;
char *mapname;
int flag;
/*
 bit0-7: mode
         0= raw mapname as set with Imgmap_set_mapname() 
         1= name of list of file names
         2= name of list of samples 
         3= name of -list_doubles cursor file
         4= name of break file
         5= name of file with server address (pubno)
*/
{
 int modus;

 modus= flag&255;
 if(modus==0) {
   strcpy(mapname,imp->mapname);
 } else if (modus==1) {
   sprintf(mapname,"%s_nam",imp->mapname);
 } else if (modus==2) {
   sprintf(mapname,"%s_smp%ld",imp->mapname,imp->spoints);
 } else if (modus==3) {
   sprintf(mapname,"%s_list_doubles_pos",imp->mapname);
 } else if (modus==4) {
   sprintf(mapname,"%s_break",imp->mapname);
 } else if (modus==5) {
   sprintf(mapname,"%s_pubno",imp->mapname);
 } else if (modus==6) {
   sprintf(mapname,"%s_exempt",imp->mapname);
 } else {
   return(0);
 }
 return(1);
}


int Imgmap_set_disabled(imp,value,flag)
struct ImgmaP *imp;
int value;
int flag;
{
 imp->disabled= !!value;
 return(1);
}


int Imgmap_get_disabled(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->disabled);
}


int Imgmap_set_use_exemptlist(imp,value,flag)
struct ImgmaP *imp;
int value;
int flag;
{
 imp->use_exemptlist= !!value;
 return(1);
}


int Imgmap_get_use_exemptlist(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->use_exemptlist);
}


int Imgmap_set_guarding(imp,value,flag)
struct ImgmaP *imp;
int value;
int flag;
{
 imp->guarding= !!value;
 return(1);
}


int Imgmap_get_guarding(imp,flag)
struct ImgmaP *imp;
int flag;
{
 return(imp->guarding);
}


int Imgmap_reset_errmsg(imp,flag)
struct ImgmaP *imp;
int flag;
{
 imp->errmsg[0]= 0;
 return(1);
}


char *Imgmap_get_errmsg(imp,msg,flag)
struct ImgmaP *imp;
char *msg;
int flag;
{
 strcpy(msg,imp->errmsg);
 return(msg);
}


int Imgmap_set_match_par(imp,search_method,
                         color_eps,bw_eps,color_max_misses,bw_max_misses,flag)
struct ImgmaP *imp;
int search_method;
double color_eps,bw_eps;
int color_max_misses,bw_max_misses;
int flag;
{
 if(search_method>=0)
   imp->search_method= search_method;
 if(color_eps>=0.0)
   imp->color_eps= color_eps;
 if(bw_eps>=0.0)
   imp->bw_eps= bw_eps;
 if(color_max_misses>=0)
   imp->color_max_misses= color_max_misses;
 if(bw_max_misses>=0)
   imp->bw_max_misses= bw_max_misses;
 return(1);
}


int Imgmap_temp_sample(imp,samp,flag)
struct ImgmaP *imp;
struct ImgsamplE **samp;
int flag;
{
 int ret;

 Imgmap_reset_errmsg(imp,0);
 ret= Imgsample_new(samp,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
           sizeof(**samp));
   return(-1);
 }
 (*samp)->spoints= imp->spoints;
 return(1);
}


int Imgmap_fopen(imp,mode,flag)
struct ImgmaP *imp;
char *mode;
int flag;
/*
 bit0-3: mode for Imgmap_get_mapname()
         1= name of list of file names
         2= name of list of samples
 bit14= create in any case
 bit15= close only, do not open new file pointer
*/
{
 char filename[SfileadrL];
 int typ;
 FILE *fp;

 if(!(flag&(1<<15))) {
   if(imp->disabled) {
disabled:;   
     Imgmap_get_mapname(imp,filename,0);
     sprintf(imp->errmsg,
             "Access to guarded map file '%s' is disabled.",filename);
     return(0);
   }
   if(!imp->guarding) {
     Imgmap_get_mapname(imp,filename,5);
     if(Sfile_size(filename,1)!=-1)
       goto disabled;
   }
 }
 typ= flag&15;
 if(typ!=1 && typ!=2)
   typ= 1;
 
 Imgmap_get_mapname(imp,filename,typ);
 if(typ==1)
   fp= imp->fp_nam;
 else
   fp= imp->fp_samp;
 if(fp!=NULL){
   fclose(fp); fp= NULL;
 }
 if(!(flag&(1<<15))){
   if(flag&(1<<14))
     if((fp= fopen(filename,"ab"))!=NULL)
       {fclose(fp); fp= NULL;}
   fp= fopen(filename,mode);
   if(fp==NULL)
     sprintf(imp->errmsg,"cannot open map file '%s' errno=%ld",filename,errno);
 }
 if(typ==1)
   imp->fp_nam= fp;
 else
   imp->fp_samp= fp;
 return(fp!=NULL);
}


int Imgmap_fclose(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0-3: mode for Imgmap_get_mapname()
         1= name of list of file names
         2= name of list of samples
*/
{
 Imgmap_fopen(imp,"",(flag&15)|(1<<15));
 return(1);
}


int Imgmap_write_nam(imp,filename,nam_adr,flag)
struct ImgmaP *imp;
char *filename;
unsigned long *nam_adr;
int flag;
/*
 bit0= address with  *nam_adr 
 bit15= write a zero
*/
{
 int ret;
 char null=0;

 if(flag&1) 
   fseek(imp->fp_nam,*nam_adr,0);
 if(nam_adr!=NULL)
   *nam_adr= ftell(imp->fp_nam);
 if(flag&(1<<15) || filename==NULL) {
   ret= fwrite(&null,1,1,imp->fp_nam);
   if(ret!=1) 
     return(0);
 }else{
   ret= fwrite(filename,strlen(filename),1,imp->fp_nam);
   if(ret!=1)
     return(0);
   fputc('\n',imp->fp_nam);
 }
 return(1);
}


int Imgmap_write_samp(imp,writesample,flag)
struct ImgmaP *imp;
struct ImgsamplE *writesample;
int flag;
/*
 bit0= address with writesample->samp_adr
 bit1= this is certainly a new address

 bit15= mark written sample as completely invalid
*/
{
 int i,ret,head,sp;
 unsigned long sadr;
 struct ImgsamplE *samp=NULL;

 if(flag&1) 
   fseek(imp->fp_samp,writesample->samp_adr,0);
 sadr= ftell(imp->fp_samp); 
 if(flag&(1<<15)) 
   writesample->invalid= -1;
 if(flag&2){
   ret= Imgmap_new_samp(imp,sadr,&samp,0);
   if(ret<=0)
     goto ex;
 }else{
   if(Imgmap_find_in_cache(imp,sadr,&samp,0)<=0) {
     ret= Imgmap_new_samp(imp,sadr,&samp,0);
     if(ret<=0)
       goto ex;
   }
 }
 Imgsample_set_samp_adr(writesample,sadr,0);
 if(samp!=NULL) {                                         /* write to cache */
   ret= Imgsample_copy(writesample,samp,0);
   if(ret<=0)
     goto ex;
 }

 head= writesample->invalid;
 if(fwrite(&head,sizeof(int),1,imp->fp_samp)!=1)
   return(0);
 if(fwrite(&(writesample->id),sizeof(writesample->id),1,imp->fp_samp)!=1) 
   return(0);
 if(fwrite(&(writesample->nam_adr),sizeof(writesample->nam_adr),1,
           imp->fp_samp)!=1) 
   return(0);
 if(fwrite(&(writesample->crc),sizeof(writesample->crc),1,imp->fp_samp)!=1) 
   return(0);
 sp= imp->spoints;
 if(fwrite(writesample->red,sp*sizeof(Imgsamp_coorD),1,imp->fp_samp)!=1)
   return(0);
 if(fwrite(writesample->green,sp*sizeof(Imgsamp_coorD),1,imp->fp_samp)!=1)
   return(0);
 if(fwrite(writesample->blue,sp*sizeof(Imgsamp_coorD),1,imp->fp_samp)!=1)
   return(0);
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_read_samp(imp,readsample,flag)
struct ImgmaP *imp;
struct ImgsamplE *readsample;
int flag;
/*
  bit0= address with readsample->samp_adr

  bit8= caching permitted
  bit9= set readsample->samp_adr according to effective address
 bit10= with caching: get filename by reference rather than by copy
*/
{
 int i,sp,ret;
 unsigned long samp_adr;
 struct ImgsamplE *samp=NULL;

 if((flag&256) && (imp->caching&3)==3){
   if(flag&1) 
     Imgmap_find_in_cache(imp,readsample->samp_adr,&samp,0);
   else
     samp= imp->searchsamp;
   if(samp==NULL)
      return(0);
 }
 if(samp!=NULL) {
   samp_adr= readsample->samp_adr;
   ret= Imgsample_copy(samp,readsample,!!(flag&1024));
   if(ret<=0)
     return(ret);
   if(!(flag&512))
     readsample->samp_adr= samp_adr;
   imp->searchsamp= samp->next;
   return(1);
 }

 Imgsample_set_filename(readsample,NULL,0);

 if(flag&1)
   fseek(imp->fp_samp,readsample->samp_adr,0);
 if(flag&512)
   readsample->samp_adr= ftell(imp->fp_samp);;
 if(fread(&(readsample->invalid),sizeof(readsample->invalid),1,imp->fp_samp)!=1)
   return(0);
 if(fread(&(readsample->id),sizeof(readsample->id),1,imp->fp_samp)!=1)
   return(0);
 if(fread(&(readsample->nam_adr),sizeof(readsample->nam_adr),1,imp->fp_samp)!=1)
   return(0);
 if(fread(&(readsample->crc),sizeof(readsample->crc),1,imp->fp_samp)!=1)
   return(0);
 sp=readsample->spoints= imp->spoints;
 if(fread(readsample->red,sp*sizeof(Imgsamp_coorD),1,imp->fp_samp)!=1)
   return(0);
 if(fread(readsample->green,sp*sizeof(Imgsamp_coorD),1,imp->fp_samp)!=1)
   return(0);
 if(fread(readsample->blue,sp*sizeof(Imgsamp_coorD),1,imp->fp_samp)!=1)
   return(0);
 return(1);
}


int Imgmap_record(imp,filename,writesample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *writesample;
int flag;
{
 int ret,i;
 char name[SfileadrL];
 unsigned long nam_adr,samp_adr;

 imp->fp_nam= NULL;
 imp->fp_samp= NULL;
 Imgmap_reset_errmsg(imp,0);

#ifndef Imgmap_dirty_fopeN

 /* filename */
 ret= Imgmap_fopen(imp,"ab",1);
 if(ret<=0)
   return(ret);
 if(Imgmap_write_nam(imp,filename,&nam_adr,0)<=0)
   goto write_err;
 Imgmap_fclose(imp,1);
 Imgsample_set_nam_adr(writesample,nam_adr,0);

 /* sample */
 ret= Imgmap_fopen(imp,"ab",2);
 if(ret<=0)
   return(ret);
 if(Imgmap_write_samp(imp,writesample,2)<=0) 
   goto write_err;
 Imgmap_fclose(imp,2);

#else /* ! Imgmap_dirty_fopeN */

 Imgmap_get_mapname(imp,name,1);
 imp->fp_nam= fopen(name,"ab");
 if(imp->fp_nam==NULL) {
   sprintf(imp->errmsg,"cannot open map file '%s' errno=%ld",name,errno);
   return(0);
 }

 if(Imgmap_write_nam(imp,filename,&nam_adr,0)<=0)
   goto write_err;
 fclose(imp->fp_nam); imp->fp_nam= NULL;
 Imgsample_set_nam_adr(writesample,nam_adr,0);

 /* sample */
 Imgmap_get_mapname(imp,name,2);
 imp->fp_samp= fopen(name,"ab");
 if(imp->fp_samp==NULL) {
   sprintf(imp->errmsg,"cannot open map file '%s' errno=%ld",name,errno);
   return(0);
 }
 if(Imgmap_write_samp(imp,writesample,2)<=0) 
   goto write_err;
 fclose(imp->fp_samp); imp->fp_samp= NULL;

#endif /* Imgmap_dirty_fopeN */

 return(1);

write_err:;
 if(imp->fp_nam!=NULL) 
   {fclose(imp->fp_nam); imp->fp_nam= NULL;}
 if(imp->fp_samp!=NULL)
   {fclose(imp->fp_samp); imp->fp_samp= NULL;}
 sprintf(imp->errmsg,"cannot write name to map file '%s' errno=%ld",
                     name,errno);
 return(0);
}


int Imgmap_truncate(imp,name,length,flag)
struct ImgmaP *imp;
char *name;
unsigned long length;
int flag;
{
 if(truncate(name,length)==-1){
   sprintf(imp->errmsg,
           "on deleting last entry: truncate(,%ld) failed, errno=%ld",
           length,errno);
   sprintf(imp->errmsg+strlen(imp->errmsg),"\nin imgmap file '%s'\n",name);
   return(0);
 }
 return(1);
}


int Imgmap_del_back(imp,filename,flag)
struct ImgmaP *imp;
char *filename;
int flag;
{
 int ret;
 struct ImgsamplE *samp= NULL;
 char lastname[SfileadrL],name[SfileadrL];

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;

 ret= Imgmap_search_adr(imp,lastname,samp,samp,1);
 if(ret<=0) {
   sprintf(imp->errmsg,"on deleting last entry: no entries found");
   goto ex;
 }
 if(strcmp(filename,lastname)!=0) {
   sprintf(imp->errmsg,"on deleting last entry: filenames differ");
   sprintf(imp->errmsg+strlen(imp->errmsg),"\ngiven '%s'",filename);
   sprintf(imp->errmsg+strlen(imp->errmsg),"\nfound '%s'",lastname);
   ret= 0; goto ex;
 }
 Imgmap_get_mapname(imp,name,2);
 if(Imgmap_truncate(imp,name,samp->samp_adr,0)<=0)
   {ret= 0; goto ex;}
 Imgmap_get_mapname(imp,name,1);
 if(Imgmap_truncate(imp,name,samp->nam_adr,0)<=0)
   {ret= 0; goto ex;}
 Imgmap_del_sample_back(imp,0);
 ret= 1;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_search_reset(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= no Imgmap_reset_errmsg()
*/
{
 if(!(flag&1))
   Imgmap_reset_errmsg(imp,0);
 imp->searchpos= 0;
 if(imp->cache!=NULL)
   imp->searchsamp= imp->cache->start_samp;
 else
   imp->searchsamp= NULL;
 return(1);
}


int Imgmap_test_diff(imp,insample,samp,max_misses,in_eps,flag)
struct ImgmaP *imp;
struct ImgsamplE *insample;
struct ImgsamplE *samp;
int max_misses;
double in_eps;
int flag;
{
 int i,misses,eps;

 eps= in_eps*Imgsamp_scalE;
 misses= 0;
 for(i=0;i<imp->spoints && misses<=max_misses;i++) {
   if(abs(insample->red[i]-samp->red[i])>eps) {
     misses++;
 continue;
   }
   if(abs(insample->green[i]-samp->green[i])>eps) {
     misses++;
 continue;
   }
   if(abs(insample->blue[i]-samp->blue[i])>eps) {
     misses++;
 continue;
   }
 }
 if(i==imp->spoints && misses<=max_misses)
   return(1);
 return(0);
}


int Imgmap_test_double_diff(imp,insample,samp,max_misses,in_eps,flag)
/* relative to center of mass : invariant against movement in color space */
struct ImgmaP *imp;
struct ImgsamplE *insample;
struct ImgsamplE *samp;
int max_misses;
double in_eps;
int flag;
{
 int i,misses;
 double eps;
 double v[Imgsamp_max_sizE][3],cm[3];

 misses= 0;
 cm[0]= cm[1]= cm[2]= 0.0;
 eps= in_eps*Imgsamp_scalE;
 for(i=0;i<imp->spoints;i++) {
   cm[0]+= 
   v[i][0]= (double)(insample->red[i]-samp->red[i]);
   cm[1]+= 
   v[i][1]= (double)(insample->green[i]-samp->green[i]);
   cm[2]+= 
   v[i][2]= (double)(insample->blue[i]-samp->blue[i]);
 }
 cm[0]/= (double) imp->spoints;
 cm[1]/= (double) imp->spoints;
 cm[2]/= (double) imp->spoints;
 for(i=0;i<imp->spoints && misses<=max_misses;i++) {
   if(fabs(v[i][0]-cm[0])>eps) {
     misses++;
 continue;
   }
   if(fabs(v[i][1]-cm[1])>eps) {
     misses++;
 continue;
   }
   if(fabs(v[i][2]-cm[2])>eps) {
     misses++;
 continue;
   }
 }
 if(i==imp->spoints && misses<=max_misses)
   return(1);
 return(0);
}


int Imgmap_test_scaled_diff(imp,insample,incm,samp,max_misses,in_eps,flag)
/* center of mass matched by three independend scale factors : 
   invariant 0-centered scaling in color space */
struct ImgmaP *imp;
struct ImgsamplE *insample;
double incm[3];
struct ImgsamplE *samp;
int max_misses;
double in_eps;
int flag;
{
 int i,misses,ret;
 double eps;
 double v[Imgsamp_max_sizE][3],cm[3],scale[3],movement[3];
 static double dark_limit= 16*Imgsamp_scalE;
 static double fixpoint[3]={ 64.0*Imgsamp_scalE,
                             64.0*Imgsamp_scalE, 64.0*Imgsamp_scalE};

 if(incm[0]<dark_limit && incm[1]<dark_limit && incm[2]<dark_limit) {
   ret= Imgmap_test_diff(imp,insample,samp,max_misses,in_eps,0);
   return(ret);
 }
 misses= 0;
 cm[0]= cm[1]= cm[2]= 0.0;
 eps= in_eps*Imgsamp_scalE;
 for(i=0;i<imp->spoints;i++) {
   cm[0]+= samp->red[i];
   cm[1]+= samp->green[i];
   cm[2]+= samp->blue[i];
 }
 cm[0]/= (double) imp->spoints;
 cm[1]/= (double) imp->spoints;
 cm[2]/= (double) imp->spoints;
 if((cm[0]-fixpoint[0])>=dark_limit)
   scale[0]= (incm[0]-fixpoint[0])/(cm[0]-fixpoint[0]);
 else
   scale[0]= 1.0;
 if((cm[1]-fixpoint[1])>=dark_limit)
   scale[1]= (incm[1]-fixpoint[1])/(cm[1]-fixpoint[1]);
 else
   scale[1]= 1.0;
 if((cm[2]-fixpoint[2])>=dark_limit)
   scale[2]= (incm[2]-fixpoint[2])/(cm[2]-fixpoint[2]);
 else
   scale[2]= 1.0;
 movement[0]= (1.0-scale[0])*fixpoint[0];
 movement[1]= (1.0-scale[1])*fixpoint[1];
 movement[2]= (1.0-scale[2])*fixpoint[2];
 for(i=0;i<imp->spoints && misses<=max_misses;i++) {
   if(fabs(((double) samp->red[i])*scale[0]+movement[0]
           -((double) insample->red[i]))>eps) {
     misses++;
 continue;
   }
   if(fabs(((double) samp->green[i])*scale[1]+movement[1]
      -((double) insample->green[i]))>eps) {
     misses++;
 continue;
   }
   if(fabs(((double) samp->blue[i])*scale[2]+movement[2]
      -((double) insample->blue[i]))>eps) {
     misses++;
 continue;
   }
 }
 if(i==imp->spoints && misses<=max_misses)
   return(1);
 return(0);
}


int Imgmap_search_adr(imp,filename,insample,foundsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *insample;
struct ImgsamplE *foundsample;
int flag;
/*
 bit0= search last entry
 bit1= accept any valid entry
 bit2= use exempt list despite imp->use_exemptlist==0
*/
{
 int ret,i,j,c,sample_existing=0,invalid_mask= Imgsamp_valid_requireD;
 int max_misses,misses,colorful= 0,boring,nearby;
 char name[SfileadrL];
 unsigned long adrmerk,nam_adrmerk,samp_adr;
 struct ImgsamplE *samp= NULL;
 double eps,d,dd,cm[3];

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;

 if(flag&1)
   Imgmap_search_reset(imp,1);

 /* use higher accuracy if insample is boring (e.g. black+white) */
 eps= imp->color_eps;
 max_misses= imp->color_max_misses;
 colorful= 0;
 boring= 0;
 if(flag&(1|2)) {
   colorful= imp->spoints;
 } else {
   cm[0]= cm[1]= cm[2]= 0.0;
   for(i=0;i<imp->spoints;i++) {
     cm[0]+= insample->red[i];
     cm[1]+= insample->green[i];
     cm[2]+= insample->blue[i];
   }
   for(j=0;j<3;j++)
     cm[j]/= ((double) imp->spoints);
   for(i=0;i<imp->spoints;i++) {
     if(imp->search_method==1) {
       d= abs(insample->red[i]-cm[0]-insample->green[i]+cm[1]);
       if(abs(insample->red[i]-cm[0]-insample->blue[i]+cm[2])>d) 
          d= abs(insample->red[i]-cm[0]-insample->blue[i]+cm[2]);
     }else{
       d= abs(insample->red[i]-insample->green[i]);
       if(abs(insample->red[i]-insample->blue[i])>d) 
          d= abs(insample->red[i]-insample->blue[i]);
     }
     if(d/Imgsamp_scalE>imp->color_eps)
       colorful++;
     nearby= 0;
     for(j=0;j<imp->spoints;j++) {
       if(j==i)
     continue;
       d= abs(insample->red[i]-insample->red[j]);
       dd= abs(insample->green[i]-insample->green[j]);
       if(dd>d)
         d= dd;
       dd= abs(insample->blue[i]-insample->blue[j]);
       if(dd>d)
         d= dd;
       if(d/Imgsamp_scalE<imp->color_eps)
         nearby++;
     }
     if(nearby >= imp->spoints/2 ) {
       boring= 1;
   break;
     }
   }
   if(colorful<imp->spoints/2 || boring) {
     eps= imp->bw_eps;
     max_misses= imp->bw_max_misses;
   }
 }

 ret= Imgmap_fopen(imp,"rb",2);
 if(ret<=0)
   goto ex;
 fseek(imp->fp_samp,imp->searchpos,0);

 while(1){
   samp_adr= ftell(imp->fp_samp);
   if(Imgmap_read_samp(imp,samp,256|512|1024)<=0)
 break;
   samp_adr= samp->samp_adr;
   if(samp->invalid&invalid_mask)
 continue;
   sample_existing= 1;
   if(flag&1) {
     ret= Imgsample_copy(samp,foundsample,0);
     if(ret<=0)
       goto ex;
 continue;
   }
   if((flag&2)){ 
     ret= 1;
   } else {
     if(imp->search_method==1) {
       ret= Imgmap_test_double_diff(imp,insample,samp,max_misses,eps,0);
     } else if(imp->search_method==2) {
       ret= Imgmap_test_scaled_diff(imp,insample,cm,samp,max_misses,eps,0);
     } else {
       ret= Imgmap_test_diff(imp,insample,samp,max_misses,eps,0);
     }
   }
   if(ret>0) {
     ret= Imgsample_copy(samp,foundsample,0);
     if(ret<=0)
       goto ex;
     imp->searchpos= ftell(imp->fp_samp);

read_filename:;
     ret= Imgmap_fopen(imp,"rb",1);
     if(ret<=0)
       goto ex;
     fseek(imp->fp_nam,foundsample->nam_adr,0);
     i=0;
     while(1){
       c= fgetc(imp->fp_nam);
       if(c==EOF) {
         sprintf(imp->errmsg,"incomplete filename in map file '%s' errno=%ld",
                 name,errno);
         ret= 0; goto ex;
       }
       if(c=='\r') 
     continue;
       if(c=='\n') c= 0;
       filename[i++]= c;
     if(c==0) break;
     }

     ret= Imgsample_set_filename(foundsample,filename,0);
     if(ret<=0)
       goto ex;
     if(((flag&4)||imp->use_exemptlist) && insample->filename!=NULL 
        && strcmp(insample->filename,filename)!=0) {
       ret= Imgexemptlist_is_exempted(imp->exemptlist,
                                      insample->filename,filename,0);
       if(ret<0)
         goto ex;
       if(ret>0)
 continue;
       /* i consider inequality to be symmetric */
       ret= Imgexemptlist_is_exempted(imp->exemptlist,
                                      filename,insample->filename,0);
       if(ret<0)
         goto ex;
       if(ret>0)
 continue;
     }

     ret= 1; goto ex;
   }
 }
 if((flag&1) && sample_existing) 
   goto read_filename;
 ret= 0;
ex:;
 if(imp->fp_samp!=NULL)
   {fclose(imp->fp_samp); imp->fp_samp= NULL;}
 if(imp->fp_nam!=NULL)
   {fclose(imp->fp_nam); imp->fp_nam= NULL;}
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_search(imp,filename,insample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *insample;
int flag;
/*
 flag : is also sent unchanged to Imgmap_search_adr()
 bit0= search last entry
 bit1= accept any valid entry
       copy found sample to insample
 bit2= use exempt list
*/
{
 int ret;
 struct ImgsamplE *samp= NULL;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;
 ret= Imgmap_search_adr(imp,filename,insample,samp,flag);
 if(ret<=0)
   goto ex;
 if(flag&2) {
   ret= Imgsample_copy(samp,insample,0);
   if(ret<=0)
     goto ex;
 }
   
 ret= 1;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_offadr(imp,filename,foundnames,foundsamples,maxfnames,flag)
struct ImgmaP *imp;
char *filename;
char **foundnames;
struct ImgsamplE **foundsamples;
int maxfnames;
int flag;
/*
 bit0= report findings
 bit1= report findings as result
 bit2= return only one filename if possible
 bit3= require identical inode
 bit4= return if exact name is found
 bit5= return 0 if no caching 
*/
{
 int count= 0,ret,wpt= -1,i,passes=1,pass;
 char line[2*SfileadrL];
 struct ImgsamplE *samp= NULL,*foundsamp,*cache_sample;

 if(flag&32)
   if(Imgmap_find_in_cache(imp,0,&cache_sample,1)<=0)
     {ret= 0; goto ex;}

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;
 ret= Imgmap_temp_sample(imp,&foundsamp,0);
 if(ret<=0)
   goto ex;

 ret= Imgmap_read_rgb(imp,filename,samp,2);
 if(ret<=0) 
    goto ex;

 if(flag&(4|8|16))
   passes= 3;
 for(pass=passes;pass>!!(flag&8);pass--){
   Imgmap_search_reset(imp,0);
   while(Imgmap_search_adr(imp,line,samp,foundsamp,0)>0){
     if(pass==3) {
       if(strcmp(filename,line)!=0)
   continue;
     } 
     for(i=0;i<=wpt;i++) 
       if(strcmp(foundnames[i],line)==0) break;
     if(i<=wpt)
   continue;
     if(!Sfile_identical(filename,line,(pass>=2)))
   continue;
     if(wpt+1<maxfnames)
       wpt++;
     if(wpt>=0) {
       strcpy(foundnames[wpt],line);
       foundsamples[wpt]= NULL;
       if(Imgmap_find_in_cache(imp,foundsamp->samp_adr,&cache_sample,0)>0)
         foundsamples[wpt]= cache_sample;
       else
         /* >>> return a copy of local sample , somehow marked for deletion */;
     }
     if(imp->icmp!=NULL) {
       if(flag&1)
         Imgcmp_stderr(imp->icmp,line,0);
       if(flag&2)
         Imgcmp_result(imp->icmp,line,0);
     }
     count++;
     if(pass==3 && (flag&16))
       {ret= 1; goto ex;}
   }
   if(count>0)
 break;
 }
 ret= count;
ex:
 Imgsample_destroy(&samp,0);
 Imgsample_destroy(&foundsamp,0);
 return(ret);
}


int Imgmap_search_filename(imp,filename,foundsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *foundsample;
int flag;
/*
 bit0= keep files open if search is successful
 bit1= open in mode r+
 bit2= search a whole directory (all data files, not subdirectories)
       Empty text finds all files without a directory
 bit3= search a whole directory tree (including subdirectories)
       Empty text finds all files
 bit4= also return invalid items
 bit5= write filename found into parameter filename
       (senseful with empty search text)
 bit6= files are already open
 <<< bit7= searching for exactly one file among whole database
       (permission to use Imgmap_offadr)
*/
{
 int ret,i,c,sample_count=0,searchnl,shortcut_successful=0;
 unsigned long samp_adr;
 char *cpt;
 char name[SfileadrL],testname[SfileadrL],searchname[SfileadrL];
 char keptname[SfileadrL],*testname_pt;
 struct ImgsamplE *samp= NULL,*foundsample_pt;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;

 strcpy(searchname,filename);
 searchnl= strlen(searchname);
 if(flag&8) {
   if(searchnl)  
     if(searchname[searchnl-1]!='/')
       strcat(searchname,"/"); 
 } else if (flag&4) {
   if(searchnl>1)
     if(searchname[searchnl-1]=='/')
       searchname[searchnl-1]= 0;
 }
 searchnl= strlen(searchname);


#ifdef Imgmap_no_filenames_in_cachE

 /* <<< only until there is no caching for filenames */
 if(flag&128) {
   testname_pt= testname;
   ret= Imgmap_offadr(imp,filename,&testname_pt,&foundsample_pt,1,4|16|32);
   if(ret<0)
     goto ex;
   if(ret>0){
     ret= Imgsample_copy(foundsample_pt,foundsample,0);
     if(ret<=0)
       goto ex;
     ret= Imgsample_set_filename(foundsample,testname_pt,0);
     if(ret<=0)
       goto ex;
     if(flag&1)
       shortcut_successful= 1;
     else
       {ret= 2; goto ex;}
   }
 }

#endif /* Imgmap_no_filenames_in_cachE */


 if(imp->fp_samp==NULL || !(flag&64)) {
   if(imp->fp_samp!=NULL) 
     {fclose(imp->fp_samp); imp->fp_samp= NULL;}
   if(flag&2)
     ret= Imgmap_fopen(imp,"r+b",2);
   else
     ret= Imgmap_fopen(imp,"rb",2);
   if(ret<=0)
     goto ex;
 }
 if(imp->fp_nam==NULL || !(flag&64)) {
   if(imp->fp_nam!=NULL) 
     {fclose(imp->fp_nam); imp->fp_nam= NULL;}
   if(flag&2)
     ret= Imgmap_fopen(imp,"r+b",1);
   else
     ret= Imgmap_fopen(imp,"rb",1);
   if(ret<=0) {
     Imgmap_fclose(imp,1);
     goto failed;
   }
 }


#ifdef Imgmap_no_filenames_in_cachE

 if(shortcut_successful)
   {ret= 2; goto ex;}

#endif /* Imgmap_no_filenames_in_cachE */


 /* ??? introduce simplified search over cache ??? */


 fseek(imp->fp_samp,imp->searchpos,0);
 while(1){
   if(Imgmap_read_samp(imp,samp,256|512|1024)<=0)
     goto done;
   if(samp->invalid&Imgsamp_valid_requireD)
 continue;

   sample_count++;
   imp->searchpos= ftell(imp->fp_samp);

   if(samp->filename!=NULL) {
     strcpy(testname,samp->filename);
   } else {
     fseek(imp->fp_nam,samp->nam_adr,0);
     if(Sfile_fgets(testname,sizeof(testname)-1,imp->fp_nam)==NULL) {
       sprintf(imp->errmsg,"cannot read filename from map file '%s' errno=%ld",
               name,errno);
       goto failed;
     }
   }
   strcpy(keptname,testname);
   if(flag&8) { 
     testname[searchnl]= 0;
   } else if(flag&4) {
     cpt= strrchr(testname,'/');
     if(cpt==NULL)
       cpt= testname;
     *cpt= 0;
   }
   if(strcmp(searchname,testname)==0) {
     if(flag&32)
       strcpy(filename,keptname);
     if(!(flag&1)){
       fclose(imp->fp_nam); imp->fp_nam= NULL;
       fclose(imp->fp_samp); imp->fp_samp= NULL;
     }
     ret= Imgsample_copy(samp,foundsample,0);
     if(ret<=0)
       goto ex;
     ret= Imgsample_set_filename(foundsample,filename,0);
     if(ret<=0)
       goto ex;
     ret= sample_count;
     goto ex;
   }
 }
done:;
failed:;
 if(imp->fp_samp!=NULL)
   {fclose(imp->fp_samp); imp->fp_samp= NULL;}
 if(imp->fp_nam!=NULL)
   {fclose(imp->fp_nam); imp->fp_nam= NULL;}
 ret= 0;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_rm_sample(imp,samp,flag)
struct ImgmaP *imp;
struct ImgsamplE *samp;
int flag;
/*
 bit0= keep files open if search is successful
*/
{
 int ret;
 
 ret= Imgmap_fopen(imp,"r+b",1);
 if(ret<=0)
   goto ex;
 ret= Imgmap_fopen(imp,"r+b",2);
 if(ret<=0)
   goto ex;
 if(Imgmap_write_samp(imp,samp,(1<<15)|1)<=0)
   {ret= 0; goto ex;}
 if(Imgmap_write_nam(imp,"",&(samp->nam_adr),(1<<15)|1)<=0)
   {ret= 0; goto ex;}
 ret= 1;
ex:
 if(!(flag&1)) {
   Imgmap_fclose(imp,"",1);
   Imgmap_fopen(imp,"",2);
 }
 return(ret);
}


int Imgmap_rm_filename(imp,filename,flag)
struct ImgmaP *imp;
char *filename;
int flag;
/*
 bit0= permission to apply empty text as filename
 bit2= remove whole directory ( rm dir/*  without subdirectories)
       Empty text finds all files without a directory
 bit3= remove a whole directory tree  ( rm -r dir )
       Empty text finds all files
*/
{
 static char name[SfileadrL];
 int counter=0,null=0,i,ret;
 double minus1= -1.0;
 struct ImgsamplE *samp= NULL;
 
 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;
 Imgmap_search_reset(imp,0);

 if(strlen(filename)==0 && !(flag&1))
   {ret= 0 ; goto ex;}
 
 while(Imgmap_search_filename(imp,filename,samp,1|2|(flag&(4|8)))>0){
   ret= Imgmap_rm_sample(imp,samp,0);
   if(ret<=0)
     goto write_err;
   counter++;
 }
 ret= counter; goto ex;
write_err:;
 if(imp->fp_nam!=NULL)
   {fclose(imp->fp_nam); imp->fp_nam= NULL;}
 if(imp->fp_samp!=NULL)
   {fclose(imp->fp_samp); imp->fp_samp= NULL;}
 sprintf(imp->errmsg,"cannot write name to map file '%s' errno=%ld",name,errno);
 ret= -counter;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_refresh_filename(imp,filename,refreshsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *refreshsample;
int flag;
/*
 bit0= permission to apply empty text as filename
 bit2= refresh whole directory ( rm dir/*  without subdirectories)
       Empty text finds all files without a directory
 bit3= refresh a whole directory tree  ( rm -r dir )
       Empty text finds all files
*/
{
 static char name[SfileadrL];
 int counter=0,null=0,i,ret,hflag,shortcut_successful= 0;
 struct ImgsamplE *samp= NULL;
 
 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0) 
   goto ex;
 Imgmap_search_reset(imp,0);
 
 if(strlen(filename)==0 && !(flag&1)) 
   {ret= 0; goto ex;}
 
 hflag= 1|2|(flag&(4|8));
 if(!(flag&(1|4|8)))
   hflag|= 128;
 while((ret= Imgmap_search_filename(imp,filename,samp,hflag))>0){
   if(ret==2)
     shortcut_successful= 1;
   refreshsample->samp_adr= samp->samp_adr;
   refreshsample->nam_adr= samp->nam_adr;
   refreshsample->invalid&= ~(1|2);               /* samp_adr, nam_adr valid */
   if(Imgmap_write_samp(imp,refreshsample,1)<=0)
     goto write_err;
   fclose(imp->fp_nam); imp->fp_nam= NULL;
   fclose(imp->fp_samp); imp->fp_samp= NULL;
   counter++;
   if(shortcut_successful)
 break;
 }
 if(counter<=0)
   sprintf(imp->errmsg,"refresh failed. adress not registered : '%s'",name);
 ret= counter; goto ex;
write_err:;
 if(imp->fp_nam!=NULL)
   {fclose(imp->fp_nam); imp->fp_nam= NULL;}
 if(imp->fp_samp!=NULL)
   {fclose(imp->fp_samp); imp->fp_samp= NULL;}
 sprintf(imp->errmsg,
         "cannot write name to map file '%s' errno=%ld",name,errno);
 ret= -counter;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_lookup(imp,filename,foundsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *foundsample;
int flag;
{
 int ret;

 Imgmap_search_reset(imp,0);
 ret= Imgmap_search_filename(imp,filename,foundsample,0);
 if(ret>0)
   ret= 1;
ex:;
 return(ret);
}


int Imgmap_copy_all(source,target,flag)
struct ImgmaP *source,*target;
int flag;
{
 int ret;
 char filename[SfileadrL],line[2*SfileadrL];
 struct ImgsamplE *samp= NULL;
 
 ret= Imgmap_temp_sample(source,&samp,0);
 if(ret<=0) 
   goto ex;
 
 Imgmap_reset_errmsg(source,0);
 Imgmap_reset_errmsg(target,0);
 Imgmap_search_reset(source,0);
 filename[0]= 0;
 while(Imgmap_search_filename(source,filename,samp,8|32)>0){
   ret= Imgmap_record(target,filename,samp,0);
   if(ret<=0)
     goto ex;
   filename[0]= 0;
 }
 if(Imgmap_get_errmsg(source,line,0)[0]!=0 || 
    Imgmap_get_errmsg(target,line,0)[0]!=0)
   ret= 0;
 else
   ret= 1;
ex:;
 Imgsample_destroy(&samp,0);
 return(ret);
}


int Imgmap_activate_exempt(imp,flag)
struct ImgmaP *imp;
int flag;
/*
 bit0= read in any case
*/
{
 int ret;
 char listname[SfileadrL];

 ret= Imgmap_get_mapname(imp,listname,6);
 if(ret<=0)
   return(ret);
 if(imp->exemptlist==NULL) {
   ret= Imgexemptlist_new(&(imp->exemptlist),listname,0);
 } else {
   /* if !bit0 : read only if file has changed */
   ret= Imgexemptlist_read(imp->exemptlist,listname,!(flag&1));
 }
 return(ret);
}


int Imgmap_add_exempt(imp,filename,lookalikes,flag)
/* add an exempt */
struct ImgmaP *imp;
char *filename;
struct LstrinG **lookalikes; /* these may get grabbed by Imgexempt_new()
                                *lookalikes is set to NULL in that case */
int flag;
{
 int ret;
 struct LstrinG *ls;
 char listname[SfileadrL];

 ret= Imgmap_get_mapname(imp,listname,6);
 if(ret<=0)
   goto ex;
 ret= Sfile_string_to_file(listname,filename,2|4);
 if(ret<=0)
   goto ex;
 for(ls= *lookalikes;ls!=NULL;ls=ls->next) {
   ret= Sfile_string_to_file(listname,ls->text,2|4);
   if(ret<=0)
     goto ex;
 }
 ret= Sfile_string_to_file(listname,"",2|4);
 if(ret<=0)
   goto ex;

 if(!Imgmap_get_use_exemptlist(imp,0))
   {ret= 1; goto ex;} 

 if(Imgexemptlist_can_still_add(imp->exemptlist,0)) {
   ret= Imgexemptlist_add(imp->exemptlist,listname,filename,lookalikes,0);
 } else {
   ret= Imgmap_activate_exempt(imp,0);
 }
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_garbage_coll(imp,flag)
struct ImgmaP *imp;
int flag;
{
 struct ImgmaP target_data,*target,buff_data,*buff;
 char mapname[SfileadrL],filename[SfileadrL],newname[SfileadrL];
 int i,file_exists=0,ret;
 struct stat stbuf;

 /* >>> should become dynamically allocated memory objects */
 target= &target_data;
 buff= &buff_data;

 Imgmap_delete_cache(imp,0);
 Imgmap_fclose(imp,1);
 Imgmap_fclose(imp,2);
 Imgmap_copy(imp,target,2|4); 
 Imgmap_get_mapname(imp,mapname,0);
 strcat(mapname,"_garb");
 Imgmap_set_mapname(target,mapname,0);
 for(i=1;i<5;i++){
   Imgmap_get_mapname(target,filename,i);
   if(stat(filename,&stbuf)!=-1) {
     if(!file_exists)
       sprintf(imp->errmsg,"cannot create temporary data base %s\n",mapname);
     sprintf(imp->errmsg+strlen(imp->errmsg),"file already existing: %s",
                                             filename);
     file_exists= 1;
   }
 }
 if(file_exists) 
   return(-1);
 ret= Imgmap_copy_all(imp,target,0);
 if(ret<=0){
   /* on failure remove partially writen files */
   for(i=1;i<5;i++){
     Imgmap_get_mapname(target,filename,i);
     if(stat(filename,&stbuf)!=-1)
       unlink(filename);
   }
   return(-1);
 }

 /* rename old files */
 Imgmap_copy(imp,buff,2|4);
 Imgmap_get_mapname(imp,mapname,0);
 strcat(mapname,"_");
 strcat(mapname,Sfile_datestr(time(0),3));
 Imgmap_set_mapname(buff,mapname,0);
 for(i=1;i<5;i++){
   Imgmap_get_mapname(imp,filename,i);
   Imgmap_get_mapname(buff,newname,i);
   if(stat(filename,&stbuf)!=-1) {
     if(rename(filename,newname)==-1){
       sprintf(imp->errmsg,"cannot rename old file %s \n",mapname);
       for(i=i-1;i>=0;i--){
         Imgmap_get_mapname(imp,filename,i);
         Imgmap_get_mapname(buff,newname,i);
         if(stat(filename,&stbuf)!=-1)
           rename(newname,filename);
       }
       Imgmap_get_mapname(target,mapname,0);
       sprintf(imp->errmsg+strlen(imp->errmsg),
               "please rename new data base %s* manually",mapname);
       return(0);
     }
   }
 }

 /* rename new files */
 for(i=1;i<5;i++){
   Imgmap_get_mapname(imp,newname,i);
   Imgmap_get_mapname(target,filename,i);
   if(stat(filename,&stbuf)!=-1)
     rename(filename,newname);
 }

 /* >>> Imgmap_deactivate(,1) : buff target  */

 if(imp->caching) {
   Imgmap_fclose(imp,1);
   Imgmap_fclose(imp,2);
   ret= Imgmap_load_cache(imp,0);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Imgmap_attach_cache(imp,cache,flag)
struct ImgmaP *imp;
struct ImgmapcachE *cache;
int flag;
{
 if(imp->cache!=NULL)
   Imgmap_delete_cache(imp,0);
 imp->cache= cache;
 if(cache!=NULL)
   cache->refcount++;
 return(1);
}


int Imgmap_make_cache(imp,flag)
struct ImgmaP *imp;
int flag;
{
 struct ImgmapcachE *cache;

 if(Imgmapcache_new(&cache,imp->mapname,0)<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
           sizeof(*cache));
   return(-1);
 }
 Imgmap_attach_cache(imp,cache,0);
 return(1);
}


int Imgmap_new_samp(imp,samp_adr,samp,flag)
struct ImgmaP *imp;
unsigned long samp_adr;
struct ImgsamplE **samp;
int flag;
{
 int ret;

 *samp= NULL;
 if((!(imp->caching&1))||(imp->cache==NULL))
   return(2);
 ret= Imgsample_new(samp,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
           sizeof(**samp));
   return(ret);
 }
 if(imp->cache->last_samp==NULL)
   imp->cache->start_samp= *samp;
 else
   imp->cache->last_samp->next= *samp;
 (*samp)->next= NULL;
 (*samp)->prev= imp->cache->last_samp;
 imp->cache->last_samp= *samp;
 (*samp)->samp_adr= samp_adr;
 return(1);
}


int Imgmap_del_sample(samp,flag)
struct ImgsamplE **samp;
int flag;
{
 struct ImgsamplE *current;

 if(*samp==NULL) return(0);
 current= (*samp)->prev;
 if(current!=NULL) 
   current->next= (*samp)->next;
 if((*samp)->next!=NULL)  
   (*samp)->next->prev= current;
 Imgsample_destroy(samp,0);
 (*samp)= current;
 return(1);
}


int Imgmap_del_sample_back(imp,flag)
struct ImgmaP *imp;
int flag;
{
 int ret;

 if((imp->caching&3)!=3) return(0);
 if(imp->cache==NULL) return(0);
 ret= Imgmap_del_sample(&(imp->cache->last_samp),0);
 if(imp->cache->last_samp==NULL)
   imp->cache->start_samp= NULL;
 imp->searchsamp= NULL;
 return(ret);
}


int Imgmap_load_cache(imp,flag)
struct ImgmaP *imp;
int flag;
{
 int ret,is_open;
 struct ImgsamplE *samp=NULL,*newsamp=NULL;
 char filename[SfileadrL],line[2*SfileadrL];
 unsigned long counter=0;

 ret= Imgmap_temp_sample(imp,&samp,0);
 if(ret<=0)
   goto ex;

 if(Imgmap_make_cache(imp,0)<=0) 
   {ret= 0; goto ex;}
 
 Imgmap_get_mapname(imp,filename,1);
 if(Sfile_size(filename,1)==-1)
   {ret= 2; goto ex;}                    /* missing database is not an error */
 ret= Imgmap_fopen(imp,"rb",1);
 if(ret<=0)
   goto ex;
 Imgmap_search_reset(imp,0);
 filename[0]= 0;
 while(Imgmap_search_filename(imp,filename,samp,1|8|32|is_open)>0){
   is_open= 64;
   if(Imgmap_new_samp(imp,samp->samp_adr,&newsamp,0)<=0) 
     {ret= 0; goto ex;}
   ret= Imgsample_copy(samp,newsamp,0);
   if(ret<=0)
     goto ex;
   filename[0]= 0;
   counter++;
   if((counter%1000)==0)
     fprintf(stderr,"loading cache : %ld\r",counter);
 }
 imp->caching|= 2;
 fprintf(stderr,"entries loaded in cache : %ld\n",counter);
 ret= 1;
ex:;
 Imgsample_destroy(&samp,0);
 Imgmap_fclose(imp,1);
 Imgmap_fclose(imp,2);
 return(ret);
}


int Imgmap_delete_cache(imp,flag)
struct ImgmaP *imp;
int flag;
{
 struct ImgsamplE *samp,*alt=NULL;
 struct ImgmapcachE *cache;
 int ret,removed=0;

 ret= Imgmapcache_destroy(&(imp->cache),0);
 removed= (ret==2);
 imp->cache= NULL;
 imp->caching&= ~2;
 return(1+removed);
}


int Imgmap_find_in_cache(imp,samp_adr,foundsample,flag)
struct ImgmaP *imp;
unsigned long samp_adr;
struct ImgsamplE **foundsample;
int flag;
/*
 bit0= do not search: return 1 if cache is valid, 0 if not
*/
{
 struct ImgsamplE *samp,*alt=NULL;

 *foundsample= NULL;
 if((!(imp->caching&1))||(imp->cache==NULL))
   return(0);
 if(flag&1)
   return(1);
 samp= imp->cache->start_samp;
 while(samp!=NULL){
   if(samp->samp_adr==samp_adr) {
     *foundsample= samp;
     return(1);
   }
   samp= samp->next;
 }
 return(0);
}


int Imgmap_open_rgb(imp,filename,rgb,flag)
struct ImgmaP *imp;
char *filename;
struct ImgrgB **rgb;
int flag;

{
 int ret;
 struct ImgstD *std= NULL;

 Imgmap_reset_errmsg(imp,0);
 Imgrgb_destroy(rgb,0);
 ret= Imgrgb_new(rgb,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
                       sizeof(*rgb));
   goto ex;
 }
 ret= Imgrgb_open(*rgb,filename,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(ret<=0 && *rgb!=NULL) {
   if(imp->errmsg[0]==0)
     sprintf(imp->errmsg,"reading '%s' , got error : %s",
             Imgrgb_get_openname(*rgb,0),Imgrgb_get_errmsg(*rgb,0));
   if(strcmp(Imgrgb_get_openname(*rgb,0),filename)!=0)
     sprintf(imp->errmsg+strlen(imp->errmsg),
             "\n(original filename '%s')",filename);
   Imgrgb_destroy(rgb,0);
 }
 return(ret);
}


int Imgmap_read_rgb(imp,filename,readsample,flag)
struct ImgmaP *imp;
char *filename;
struct ImgsamplE *readsample;
int flag;
/*
 bit0= do not check for serialized sample data in filename
 bit1= remove sample data from filename if it is a serialized sample
*/
{
 int ret,status;
 struct ImgrgB *rgb= NULL;
 struct ImgstD *std= NULL;

 Imgmap_reset_errmsg(imp,0);
 if(!(flag&1)) {
   ret= Imgsample_deserialize(readsample,filename,&status,0);
   if(ret>0) {      /* this is really a serialized sample and no fileaddress */
     if(flag&2) 
       strcpy(filename,readsample->filename);
     goto ex;
   }
 }
 ret= Imgmap_open_rgb(imp,filename,&rgb,0);
 if(ret<=0)
   goto ex;
 ret= Imgstd_new(&std,rgb,imp->spoints,0);
 if(ret<=0) {
   sprintf(imp->errmsg,"cannot create memory object with %ld bytes",
                       sizeof(*std,0));
   goto ex;
 }
 ret= Imgstd_read_rgb(std,imp->errmsg,0);
 if(ret<=0)
   goto ex;
 ret= Imgsample_set_rgb(readsample,imp->spoints,
                        std->red,std->green,std->blue,0);
 if(ret<=0)
   goto ex;
 ret= Imgsample_set_filename(readsample,filename,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(ret<=0 && rgb!=NULL) {
   if(imp->errmsg[0]==0)
     sprintf(imp->errmsg,"reading '%s' , got error : %s",
                          Imgrgb_get_openname(rgb,0),Imgrgb_get_errmsg(rgb,0));
   if(strcmp(Imgrgb_get_openname(rgb,0),filename)!=0)
     sprintf(imp->errmsg+strlen(imp->errmsg),
             "\n(original filename '%s')",filename);
 }
 Imgstd_destroy(&std,0);
 Imgrgb_destroy(&rgb,0);
 return(ret);
}


