
/*
 cc -g -c imgsamp.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "imgrgb.h"

#include "imgsamp.h"



/* ------------------------------- ImgsamplE ------------------------------ */


int Imgsample_new(sample,flag)
struct ImgsamplE **sample;
int flag;
{
 int i;
 struct ImgsamplE *s;

 *sample= s= TSOB_FELD(struct ImgsamplE,1);
 if(s==NULL)
   return(-1);
 s->invalid= 15;
 s->id= 0;
 s->samp_adr= 0;
 s->nam_adr= 0;
 s->filename= NULL;
 s->filename_is_reference= 0;
 s->crc= 0;
 s->spoints= 0;
 for(i=0;i<Imgsamp_max_sizE;i++) 
   s->red[i]= s->green[i]= s->blue[i]= 0;
 s->prev= s->next= NULL;
 return(1);
}


int Imgsample_destroy(sample,flag)
struct ImgsamplE **sample;
int flag;
{
 if(*sample==NULL)
   return(0);
 if((*sample)->filename!=NULL && !(*sample)->filename_is_reference)
   Sregex_string(&((*sample)->filename),NULL,0);
 Smem_freE((char *) *sample);
 *sample= NULL;
 return(1);
}


int Imgsample_copy(source,target,flag)
struct ImgsamplE *source;
struct ImgsamplE *target;
int flag;
/*
 bit0= establish reference link to filename rather than copying it
*/
{
 int i,ret;

 target->invalid= source->invalid;
 target->id= source->id;
 target->samp_adr= source->samp_adr;
 target->nam_adr= source->nam_adr;
 ret= Imgsample_set_filename(target,source->filename,flag&1);
 if(ret<=0)
   return(ret);
 target->crc= source->crc;
 target->spoints= source->spoints;
 for(i=0;i<Imgsamp_max_sizE;i++) {
   target->red[i]= source->red[i];
   target->green[i]= source->green[i];
   target->blue[i]= source->blue[i];
 }
 return(ret);
}


int Imgsample__set_coord(coord,value,flag)
Imgsamp_coorD *coord;
double value;
int flag;
{
 if(value>=Imgsamp_scalE)
   *coord= Imgsample_max_coorD;
 else if(value<0.0)
   *coord= 0;
 else
  *coord= value*Imgsamp_scalE;
 return(1);
}


double Imgsample_get_coord(sample,color,idx,flag)
struct ImgsamplE *sample;
int color; /* 0=red , 1=green , 2=blue */
int idx;
int flag;
{
 if(idx>=sample->spoints || idx<0)
   return(0.0);
 switch (color) {
 case 0: return(((double) sample->red[idx])/((double) Imgsamp_scalE));
 case 1: return(((double) sample->green[idx])/((double) Imgsamp_scalE));
 case 2: return(((double) sample->blue[idx])/((double) Imgsamp_scalE));
 }
 return(0.0);
}


int Imgsample_set_rgb(sample,spoints,red,green,blue,flag)
struct ImgsamplE *sample;
int spoints;
double *red,*green,*blue;
int flag;
{
 int i;

 if(spoints>Imgsamp_max_sizE)
   return(0);
 sample->spoints= spoints;
 for(i=0;i<spoints;i++){
   Imgsample__set_coord(&(sample->red[i]),red[i],0);
   Imgsample__set_coord(&(sample->green[i]),green[i],0);
   Imgsample__set_coord(&(sample->blue[i]),blue[i],0);
 }
 sample->invalid&= ~8;
 return(1);
}


int Imgsample_set_nam_adr(sample,nam_adr,flag)
struct ImgsamplE *sample;
unsigned long nam_adr;
int flag;
{
 sample->nam_adr= nam_adr;
 sample->invalid&= ~2;
 return(1);
}


int Imgsample_set_samp_adr(sample,samp_adr,flag)
struct ImgsamplE *sample;
unsigned long samp_adr;
int flag;
{
 sample->samp_adr= samp_adr;
 sample->invalid&= ~1;
 return(1);
}


int Imgsample_set_filename(sample,filename,flag)
struct ImgsamplE *sample;
char *filename;
int flag;
/*
 bit0= establish reference link to filename rather than copying it
*/
{
 if(flag&1) {
   if(!sample->filename_is_reference)
     Sregex_string(&(sample->filename),NULL,0);
   sample->filename= filename;
   sample->filename_is_reference= 1;
 } else {
   sample->filename_is_reference= 0;
   if(Sregex_string(&(sample->filename),filename,0)<=0)
     return(-1);
 }
 return(1);
}


int Imgsample_serialize(sample,text,flag)
struct ImgsamplE *sample;
char **text;
int flag;
{
 char *format= NULL;
 int ret, size, i;

 size= 20+20+sample->spoints*13+100;
 format= Smem_malloC(size);
 if(format==NULL)
   {ret= -1; goto ex;} 
 Sregex_string(text,NULL,0);
 sprintf(format,"-smp%ld-%8.8X",sample->spoints,sample->invalid|3);
 sprintf(format+strlen(format),"-%8.8X-%8.8X",sample->id,sample->crc);
 for(i=0;i<sample->spoints;i++) {
   sprintf(format+strlen(format),"-%4.4X%4.4X%4.4X",
                              sample->red[i],sample->green[i],sample->blue[i]);
 }
 *text= format;
 format= NULL;
 if(Sregex_string(text,"+",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(text,sample->filename,1)<=0)
   {ret= -1; goto ex;}
 ret= 1;
ex:;
 if(format!=NULL)
   Smem_freE(format);
 return(ret);
}


int Imgsample__read_hex(rpt,value,length,flag)
char **rpt;
unsigned long value;
int length;
int flag;
/*
 bit0= expect preceding '-'
*/
{
 int ret;

 if(flag&1) {
   if(**rpt!='-')
     return(0);
   (*rpt)++;
 }
 ret= Sregex_hex_value(*rpt,value,length,0);
 if(ret<=0)
   return(ret);
 (*rpt)+= length;
 return(1);
}


int Imgsample_deserialize(sample,text,status,flag)
struct ImgsamplE *sample;
char *text;
int *status;
/*
  0= successfully completed
  1= trying to determine format
  2= trying to confirm format by reading 'invalid' , 'id' , 'crc'
  3= trying to read sample values
  4= trying to read file address
*/
int flag;
{
 int ret,i;
 unsigned long value;
 char expected[80],*rpt;

 *status= 1;
 rpt= text;
 sprintf(expected,"-smp%ld",sample->spoints);
 if(strncmp(expected,text,strlen(expected))!=0)
   {ret= 0; goto ex;}
 rpt+= strlen(expected);
 *status= 2;
 ret= Imgsample__read_hex(&rpt,&(sample->invalid),8,1);
 if(ret<=0)
   goto ex;
 ret= Imgsample__read_hex(&rpt,&(sample->id),8,1);
 if(ret<=0)
   goto ex;
 ret= Imgsample__read_hex(&rpt,&(sample->crc),8,1);
 if(ret<=0)
   goto ex;
 *status= 3;
 for(i=0;i<sample->spoints;i++) {
   ret= Imgsample__read_hex(&rpt,&value,4,1);
   if(ret<=0)
     goto ex;
   sample->red[i]= value;
   ret= Imgsample__read_hex(&rpt,&value,4,0);
   if(ret<=0)
     goto ex;
   sample->green[i]= value;
   ret= Imgsample__read_hex(&rpt,&value,4,0);
   if(ret<=0)
     goto ex;
   sample->blue[i]= value;
 }
 *status= 4;
 if(*rpt!='+')
   {ret= 0; goto ex;}
 ret= Imgsample_set_filename(sample,rpt+1,0);
 if(ret<=0)
   goto ex;
 *status= 0;
 ret= 1;
ex:;
 return(ret);
}
