

/*
 cc -g -c imvmapp.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "../s_tools/smem.h"
#include "imvguard.h"
#include "imvmapper.h"
#include "../as/asagent.h"
#include "imvmover.h"

/* <<< to be removed when helptext generation is redesigned */
#include "../as/asview.h"
#include "../as/asqueue.h"
#include "../as/asinput.h"
#include "../as/asprog.h"


#include "imvapp.h"


/* -------------------------------- ImvapP ------------------------------ */


int Imvapp_new(app,progname,flag)
struct ImvapP **app;
char *progname;
int flag;
{
 struct ImvapP *m;
 struct AsagenT *agent;
 char *cpt,line[SfileadrL];
 int i,ret;
 int Imvapp_interpreter(),Imvapp_eval(),Imvapp_determine_current();
 int Imvapp_cursor_advance(),Imvapp_seems_known();

 *app= m= TSOB_FELD(struct ImvapP,1);
 if(m==NULL)
   return(-1);
 strcpy(m->progname,progname);
 m->recursion_level= 0;
 m->recursion_max= 1000;
 m->start_dir[0]= 0;
 m->agent= NULL;
 m->guard= NULL;
 m->mapper_db= NULL;
 m->search_doubles= 0;
 m->mover= NULL;

 cpt= getcwd(m->start_dir,sizeof(m->start_dir));
 if(cpt==NULL){
   sprintf(line,"Error %ld on determining working directory :\n%s",
                errno,strerror(errno));
   goto failed;
 }

 ret= Asagent_new(&(m->agent),progname,1,0);
 if(ret<=0)
   goto failed;
 ret= Asagent_default_kbd(m->agent,0);
 if(ret<=0)
   goto failed;
 Asagent_attach_app(m->agent,(char *) m,
                     Imvapp_interpreter,Imvapp_eval,Imvapp_determine_current,
                     Imvapp_cursor_advance,Imvapp_seems_known,0);

 ret= Imvguard_new(&(m->guard),m->agent,0);
 if(ret<=0)
   goto failed;
 ret= Imvmover_new(&(m->mover),m->agent,0);
 if(ret<=0)
   goto failed;

 return(1);
failed:;
 Imvapp_destroy(app,0);
 return(ret);
}


int Imvapp_destroy(app,flag)
struct ImvapP **app;
int flag;
{
 int i;

 if(*app==NULL)
   return(0);
 Imvguard_destroy(&((*app)->guard),0);
 Sregex_string(&((*app)->mapper_db),NULL,0);
 Imvmover_destroy(&((*app)->mover),0);
 Asagent_destroy(&((*app)->agent),0);

 Smem_freE((char *) *app);
 *app= NULL;
 return(1);
}


int Imvapp_set_trashdir(app,dir,flag)
struct ImvapP *app;
char *dir;
int flag;
/*
 bit0= append
*/
{
 int ret;
 
 ret= Asagent_set_trashdir(app->agent,dir,flag);
 return(ret);
}


int Imvapp_set_maindir(app,dir,flag)
struct ImvapP *app;
char *dir;
int flag;
/*
 bit0= append
*/
{
 int ret;
 
 ret= Asagent_set_maindir(app->agent,dir,flag);
 return(ret);
}


int Imvapp_get_maindir(app,dir,flag)
struct ImvapP *app;
char **dir;
int flag;
{
 int ret;
 
 ret= Asagent_get_maindir(app->agent,dir,flag);
 return(ret);
}


int Imvapp_get_trashdir(app,dir,flag)
struct ImvapP *app;
char **dir;
int flag;
{
 int ret;
 
 ret= Asagent_get_trashdir(app->agent,dir,flag);
 return(ret);
}


int Imvapp_determine_current(app,adrpt,idx,count,flag)
struct ImvapP *app;
char **adrpt;
int *idx;
int *count;
int flag;
{
 int ret;

 ret= Imvmover_determine_current(app->mover,idx,count,0);
 *adrpt= app->mover->current_adr;
 return(ret);
}


int Imvapp_cursor_advance(app,step,flag)
struct ImvapP *app;
int *step;
int flag;
/*
 bit0= *step is the absolute index
 bit1= use *step for setting rather than for getting cursor movement
*/
{
 int ret,count;
 char *adrpt;

 if(flag&2) 
   app->mover->fstep= *step;
 else
   *step= app->mover->fstep;
 ret= Imvmover_cursor_advance(app->mover,flag&1);
 if(ret<=0)
   goto ex;
 if((flag&3)==1) 
   ret= Imvapp_determine_current(app,&adrpt,step,&count,0);
ex:;
 return(ret);
}


int Imvapp_seems_known(app,adr,filec,filev,flag)
struct ImvapP *app;
char *adr;
int *filec;
char ***filev;
int flag;
/*
 bit0= use opened reg_adr
 bit1= search even if in guarded location
*/
{
 int ret;

 ret= Imvguard_seems_known(app->guard,adr,filec,filev,flag);
 return(ret);
}


int Imvapp_hide(app,flag)
struct ImvapP *app;
int flag;
/*
 bit0= hide current
 bit1= hide doubles
 bit2= only hide views of type "child"
*/
{
 int ret;

 ret= Asagent_hide(app->agent,flag);
 return(ret);
}


int Imvapp_get_first_double(app,adr,reply,flag)
struct ImvapP *app;
char *adr;
char **reply;
int flag;
{
 int ret;
 char **filev= NULL;
 int filec= 0;

 /* >>> try to get the double list from double_viewer */;

 ret= Imvguard_seems_known(app->guard,adr,&filec,&filev,0);
 if(ret<0)
   goto ex;
 if(ret>0 && filec>0){
   *reply= filev[0];
   filev[0]= NULL; /* to protect *reply from being destroyed */
   Sfile_destroy_argv(&filec,&filev,0);
 } else {
   Asagent_stderr(app->agent,"\n--- No duplicate files found\n\n",1);
   ret= 0;
   goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Imvapp_eval_add_result(app,result_pt,text,flag)
struct ImvapP *app;
char **result_pt;
char *text;
int flag;
/*
 bit0= escape as shell parameter (assumed to be outside of quotes)
*/
{
 char esc_name[3*SfileadrL];

 if(flag&1) {
   Sfile_esc_shell(text,esc_name,0);
   if(Sregex_string(result_pt,esc_name,1)<=0)
     return(-1);
 } else {
   if(Sregex_string(result_pt,text,1)<=0)
     return(-1);
 }
 return(1);
}


int Imvapp_eval(app,in_text,result_pt,flag)
struct ImvapP *app;
char *in_text;
char **result_pt;
int flag;
/*
 bit0= evaluation for shell purposes
*/
{
 int i,l,ret,closing_bracket,idx,count;
 char *text= NULL,*result= NULL,*subresult= NULL,*start,*insert= NULL,*adr;
 char idx_text[81],dir[SfileadrL],filename[SfileadrL]; 

 if(Sregex_string(&text,in_text,0)<0)
   {ret= -1; goto ex;}
 start= text;
 l= strlen(text);
 for(i=0;i<l-1;i++) {
   if(text[i]!='$') 
 continue;
   text[i]= 0;
   if(Sregex_string(&result,start,1)<=0)
     {ret= -1; goto ex;}
   if(text[i+1]=='(') {
     closing_bracket= Sregex_closing_bracket(text+i+2,')',0);
     if(closing_bracket<0)
 break;
     text[i+2+closing_bracket]= 0;
     ret= Asagent_get_permission(app->agent,"shell",text+i+2,0);
     if(ret<=0)
       goto ex;
     ret= Imvapp_eval(app,text+i+2,&subresult,1);
     if(ret<=0)
       goto ex;
     ret= Sfile_popen_resultstring(subresult,&insert,1);
     if(ret<=0)
       goto ex;
     ret= Imvapp_eval_add_result(app,&result,insert,0);
     if(ret<=0)
       goto ex;
     Sregex_string(&insert,NULL,0);
     /* resume after closing bracket */
     i= i+2+closing_bracket;
     start= text+i+1;
   } else if(text[i+1]=='-') {
     ret= Imvapp_eval_add_result(app,&result,app->agent->trashdir,flag&1);
     if(ret<=0)
       goto ex;
     i++;
     start= text+i+1;
   } else if(text[i+1]=='*') {
     ret= Imvapp_eval_add_result(app,&result,app->mover->current_adr,
                                 flag&1);
     if(ret<=0)
       goto ex;
     i++;
     start= text+i+1;
   } else if(text[i+1]=='#') {
     ret= Imvapp_determine_current(app,&adr,&idx,&count,0);
     if(ret<=0)
       goto ex;
     sprintf(idx_text,"%ld",idx+1);
     ret= Imvapp_eval_add_result(app,&result,idx_text,flag&1);
     if(ret<=0)
       goto ex;
     i++;
     start= text+i+1;
   } else if(text[i+1]=='~') {
     ret= Sfile_home_adr("",dir,0);
     if(ret<=0) {
       Asagent_stderr(app->agent,"--- cannot determine $HOME directory",1);
       goto ex;
     }
     ret= Imvapp_eval_add_result(app,&result,dir,flag&1);
     if(ret<=0)
       goto ex;
     i++;
     start= text+i+1;
   } else if(text[i+1]=='0') {
     Sfile_split_adr(app->agent->progname,dir,filename,1);
     ret= Imvapp_eval_add_result(app,&result,dir,flag&1);
     if(ret<=0)
       goto ex;
     i++;
     start= text+i+1;
   } else if(text[i+1]=='.') {
     ret= Imvapp_eval_add_result(app,&result,app->agent->maindir,flag&1);
     if(ret<=0)
       goto ex;
     i++;
     start= text+i+1;
   } else if(text[i+1]=='$') {
     ret= Imvapp_eval_add_result(app,&result,"$",0);
     if(ret<=0)
       goto ex;
     i++;
     start= text+i+1;
   } 
 }  
 if(Sregex_string(&result,start,1)<=0)
   {ret= -1; goto ex;}
 ret= 1;
ex:;
 Sregex_string(result_pt,NULL,0);
 if(ret>0)
   *result_pt= result;
 else
   Sregex_string(&result,NULL,0);
 Sregex_string(&text,NULL,0);
 Sregex_string(&subresult,NULL,0);
 Sregex_string(&insert,NULL,0);
 return(ret);
}


int Imvapp_helptext(app,reply,flag)
struct ImvapP *app;
char **reply;
int flag;
/*
 bit0-3 = mode : 0=intro , 1=command reference , 2=appendice ,
                 3=external helpers ,  not here: 4=key map help text
*/
{
 static Asagent_helplinE intro_text[]= {
"",
"This software moves and manages image files in JPEG or GIF format.",
"Its main purpose is to show the images and to coordinate usual file",
"management tasks with a picture recognition system.",
"A list of external helper programs is displayed as appendix to this text.",
"",
"The program maintains a list of file addresses. A list index points to the",
"current item which can be manipulated. At start the list is empty. It is",
"usually filled by the program's arguments but may be extended by dialog",
"commands as well.",
"",
"Arguments are commands which get executed before the dialog starts.",
"Arguments which do not begin with a '-' character get prepended by -addfile:",
"automatically, so they are added to the list as fileadresses.",
"At the beginning of the dialog, the first item of the list is made current.",
"",
"If a file gets moved in or out a directory which contains a file with the",
"name  .imv_guarded_directory  then the picture recognition system gets",
"informed about the new address. Therefore all directories which contain",
"files known to this system should also contain a file .imv_guarded_directory",
"",
"Files that are in a directory without .imv_guarded_directory  are tested",
"by the recognition system wether they may already be stored in a guarded",
"directory. If similar files are found then they are displayed with a second",
"viewer window.",
"",
"Many commands which expect arguments like filenames or shell commands",
"perform parameter substition. This means that before execution certain text",
"parts get replaced by values known to the program. All these special",
"text parts begin with a $-sign. The next character decides about the kind",
"of substitution. See a list what is replaced by what :",
" $*       the address of the current item",
" $.       main directory (defined by -keyset:maindir)",
" $-       trash directory (defined by -keyset:trashdir)",
" $~       the user's $HOME directory",
" $0       the program's start directory (directory part of argv[0])",
" $(text)  perform parameter substitution on text, execute the result as",
"          shell command and use the standard output of this command.",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};
 static Asagent_helplinE reference_text[]= {
"",
"                Command reference of application commands",
"",
"-addfile:fileaddress",
"     Append the given file address to the list and make it the current item.",
"",
"-addfile_list:fileaddress",
"     Read the given file and append all non-empty lines as file addresses to",
"     the list. The last added item becomes the current one.",
"",
"-back",
"     Make the previous list item current.",
"",
"-doubleto:directoryaddress",
"     Move the first item of the displayed similar files to the given",
"     directory. This is mainly used to replace poor versions of an image",
"     by a new and better one.",
"     Performs parameter substition.",
"",
"-for:[+|-]start_index:[+|-]end_index:command",
"     With each list item beginning with start_index and ending with end_index",
"     perform the given command. Abort loop on error. If end_index is smaller",
"     than start_index, then command is not performed at all.",
"     The current item index is set according to the last run of command",
"     within the loop.",
"",
"-jpgb",
"     Convert the current item to old fashioned JPEG format. A quality",
"     parameter can be stored in the file .imv_jpgb_quality_value but a",
"     reasonable value is already preset at program start.",
"     The original file is preserved within the trash directory so this",
"     command does not immediately save disk space when reducing file size.",
"",
"-jump:[+|-]index",
"     Make the list item current which has the given list index.",
"     The index may be any number larger than 0. If it is preceded by a '+'",
"     or '-' then the index is added resp. subtracted from the current index.",
"     Note: -jump:+1 may cause the program to end if -end_on:list_end is set.",
"",
"-list:[+|-]start_index:[+|-]end_index[:option]",
"     List the indice and addresses of the files in the range from",
"     start_index to end_index.",
"     An index may be any number larger than 0. If it is preceded by a '+'",
"     or '-' then the index is added resp. subtracted from the current index.",
"     If option is \"raw\" then only the file addresses are listed.",
"     If option is \"count\" then also the total number of files is listed.",
"",
"-mapper_cmd:mapper_command:databaseaddress:glue_character",
"     Set the shell command, the database and the glue character (e.g. ':')",
"     for calling the picture recognition system. A call will be formed by",
"      <mapper_command><glue_character><databaseaddress> \\",
"        <option><glue_character><filename>",
"     Where option is in {-append, -refresh, -del_file, -search, -lookup_adr}",
"     This command should be used with great care, if ever. Under normal",
"     circumstances, -mapper_db  should be sufficient for all purposes.",
"",
"-mapper_db:databaseaddress",
"     Set the database to be used by the picture recognition system.",
"",
"-moveto:directoryaddress",
"     Move the current item to the given directory.",
"     Performs parameter substition.",
"",
"-preserve",
"     Make a backup copy of the current item in trashdir and enable it to be",
"     re-installed by -undo . This is useful before -shell commands which",
"     may alter the content of the current item.",
"",
"-replace_double",
"     Overwrites the first item of the displayed similar files by the current",
"     item. The duplicate's filename, its permissions and its linkage within",
"     the filesystem are preserved. Any link to the old duplicate will",
"     therefore point to the new data.",
"",
"-skip",
"     Make the next list item current.",
"     Note: -skip may cause the program to end if -end_on:list_end is set.",
"",
"-undo",
"     Revoke the most recent operation which altered the content or the",
"     address of a list item. Currently operations can be undone back",
"     to the program's start. -undo does not revoke -addfile (i.e it does",
"     not remove items from the list).",
"",
"-visit:fileaddress",
"     Append fileaddress temporarily to the file list, show the image like",
"     the current item is shown with -refresh, and then remove it from the",
"     file list.",
"     Note the special permission handling of -visit :",
"     Class is \"visit\" and action is the fileaddress. This address is",
"     checked wether it begins with the permitted action text. The remainder",
"     may not contain any \"/\" characters. If not all addresses are permitted",
"     then symbolic links or file objects with multiple hardlinks are denied.",
"     Example: -permission:all:visit:deny",
"              -permission:all:visit-/tmp/php:permit",
"              allows -visit:/tmp/phpXyZaBC but not -visit:/my/private/file",
"",
"-where:criterion:regular_expression:command",
"     For all items in the list, test the criterion with the given regular",
"     expression. If it matches, execute command with that item.",
"     Criterion may be one of :",
"       \"name\" the last part of the current item's address",
"       \"dir\"  the other part of the current item's address",
"       \"adr\"  complete current item's address",
"     For a description of regular expressions see: man 1 ed , man 7 regex",
"     The current item index is set according to the last run of command",
"     within the loop.",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};
 static Asagent_helplinE appendice_text[]= {
"",
"",
"                      Appendix application A :  Keyset Definitions",
"",
"Usually a keyset definition is written to a file and read by -keyset:readfrom",
"If the content of file /home/simv/anikeys looks like that:",
"",
"  # Clear old keyset, set main directory and trashdirectory",
"    clearall",
"    maindir:/home/pictures",
"    trashdir:$./removed_pictures",
"",
"  # Define three move targets with capital and small letters",
"    helpstart:Mice Horses Cats             (caps reform file before moving)",
"  # Map the capital letters MHC to a conversion to standard format",
"  # (they get extended below so they finally perform -jpgb before -moveto.)", 
"    map:MHC@   -jpgb",
"  # Map the keys 'M' or 'm' to : -moveto /home/pictures/mice",
"    map:Mm@    -moveto:$./mice",
"  # Map the keys 'H' or 'h' to : -moveto /home/pictures/horses",
"    map:Hh@    -moveto:$./horses",
"  # Map the keys 'C' or 'c' to : -moveto /home/pictures/cats",
"    map:Cc@    -moveto:$./cats",
"",
"",
"  # Some general keys not specific to the file management task",
"    helpmore: >forward <backward -delete ~deldouble &undo !reform",
"  # The help key",
"    map:?@     -help_short",
"  # Navigation keys : '>' and SPACE move forward",
"    map:> @    -skip",
"  # Navigation keys : '<' and BACKSPACE move backward",
"    map:<@     -back",
"    map:@#127@ -back",
"  # Deletion keys for current item and its first lookalike",
"    map:-@     -moveto:$./removed_pictures",
"    map:~@     -doubleto:$./removed_pictures",
"  # Undo and file format normalizer",
"    map:&@     -undo",
"    map:!@     -jpgb",
"",
"then the command",
"  -keyset:readfrom:/home/simv/anikeys",
"activates this definition and overwrites the old one. It does not matter",
"wether given as argument, in dialog or read from a file.",
"",
"",
"              Appendix application L : Legal stuff",
"",
"This software is copyright 2001, Thomas Schmitt stic-source@gmx.net and",
"provided to you without any warranty under an open source BSD license.", 
"(see file COPYING)",
"",
"",
"              Appendix application P : Permission Classes",
"",
"This is a list of application permission classes and their commands.",
"see \"Appendix agent P\" for more information about permission classes.",
"",
"  Class       Command (resp. Action)",
"------------------------------------------------------------------------",
"",
"  security    -mapper_cmd",
"",
"  service     -mapper_db",
"",
"  modify      -repair_icv , -jpgb",
"  move        -doubleto , -moveto , -replace_double",
"  filelist    -addfile",
"  visit       -(start text of fileaddress to visit. See above: -visit)",
"  undo        -undo , -preserve",
"",
"  navigation  -back , -skip, -jump",
"  info        -list, -refresh",
"  loop        -for , -where",
"",
"------------------------------------------------------------------------",
"",
"@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
};

 Asagent_helplinE *helptext;
 int mode,ret,i;
 char line[SfileadrL];

 if(Sregex_string(reply,"",0)<=0)
   {ret= -1; goto ex;};
 mode= flag&15;
 if(mode==0)
   helptext= intro_text;
 else if(mode==1)
   helptext= reference_text;
 else if(mode==2)
   helptext= appendice_text;
 else if(mode==3)
   goto external_helpers;
 else
   {ret= 0; goto ex;}
 for(i=0;strcmp(helptext[i],
                "@END_OF_HELPTEXT_(HOPEFULLY_A_UNIQUELY_SILLY_TEXT)@"
                )!=0;i++) {
   sprintf(line,"%s\n",helptext[i]);
   if(Sregex_string(reply,line,1)<=0)
     {ret= -1; goto ex;}
 }
 ret= 1; goto ex;

external_helpers:;
 if(Sregex_string(reply,"Image conversion program   :  ",0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(reply,"convert -comment '' -quality ",1)<=0)
   {ret= -1; goto ex;}
 sprintf(line,"%ld",Imvmover_jpgb_convert(app->mover,1));
 if(Sregex_string(reply,line,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(reply,"\nPicture recognition system :  ",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(reply,app->guard->mapper->cmd,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(reply,"\n",1)<=0)
   {ret= -1; goto ex;}
 ret= 1; goto ex;

 ret= 1;
ex:;
 return(ret);
}


int Imvapp_cmd_help(app,flag)
struct ImvapP *app;
int flag;
/*
 bit0= -help_short : display only Imvprog help text
*/
{
 int ret,i;
 char line[4096],*helptext= NULL;

 ret= Asagent_get_permission(app->agent,"help","help",0);
 if(ret<=0)
   goto ex;

 if(flag&1)
   goto help_short;

 Asagent_stdout(app->agent,"",1);
 Asagent_stdout(app->agent,"",1);
 Asagent_stdout(app->agent,"                Features and Options of ",0);
 Asagent_stdout(app->agent,app->agent->progname,1);

 for(i=0;i<3;i++) {
   if(i<2)
     ret= Imvapp_helptext(app,&helptext,i);
   else
     ret= Asagent_helptext(app->agent,&helptext,i);
   if(ret<=0)
     goto ex;
   Asagent_stdout(app->agent,helptext,4);
   if(i<2)
     ret= Asagent_helptext(app->agent,&helptext,i);
   else
     ret= Imvapp_helptext(app,&helptext,i);
   if(ret<=0)
     goto ex;
   Asagent_stdout(app->agent,helptext,4);
 }
 Asagent_stdout(app->agent,"",1);
 Asagent_stdout(app->agent,
"----------------------------- External helper programs ------------------------",
                 1);
 Asagent_stdout(app->agent,"",1);
 ret= Imvapp_helptext(app,&helptext,3);
 if(ret<=0)
   goto ex;
 Asagent_stdout(app->agent,helptext,0);
 ret= Asagent_helptext(app->agent,&helptext,3);
 if(ret<=0)
   goto ex;
 Asagent_stdout(app->agent,helptext,0);
 Asagent_stdout(app->agent,"",1);
 Asagent_stdout_bar(app->agent,1);
help_short:;
 ret= Asagent_helptext(app->agent,&helptext,4);
 if(ret<=0)
   goto ex;
 Asagent_stdout(app->agent,"",1);
 Asagent_stdout_bar(app->agent,1);
 Asagent_stdout(app->agent,helptext,1);
 Asagent_stdout_bar(app->agent,1);
 ret= 1;
ex:
 Sregex_string(&helptext,NULL,0);
 return(ret);
}


int Imvapp_cmd_mapper_cmd(app,in_cmd,flag)
struct ImvapP *app;
char *in_cmd;
int flag;
{
 int ret;
 char *cpt,*cmd= NULL,*db= NULL,glue_char= ' ';

 ret= Asagent_get_permission(app->agent,"security","mapper_cmd",0);
 if(ret<=0)
   goto ex;
 ret= Asagent_get_permission(app->agent,"service","mapper_db",0);
 if(ret<=0)
   goto ex;

 /* cmd:db:glue_char */
 cpt= in_cmd;
 ret= Sregex_cut_neck(cpt,':',&cmd,&cpt,0);
 if(ret<0)
   goto ex;
 ret= Sregex_cut_neck(cpt,':',&db,&cpt,0);
 if(ret<0)
   goto ex;
 if(*cpt!=0)
   glue_char= *cpt;
 ret= Imvguard_set_mapper_par(app->guard,cmd,db,glue_char,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 Sregex_string(&cmd,NULL,0);
 Sregex_string(&db,NULL,0);
 return(ret);
}


int Imvapp_interpreter(app,flag)
struct ImvapP *app;
int flag;
/*
 bit0= quiet : do not report command
 bit1= with bit0 : do not report user
 bit2= client mode (call Asagent_client_interpreter() and return)
 bit3= forced local mode
*/
/*
 return values: 
  <=0 = error
  1 = all went normal
  2 = redo user interaction
  3 = end agent service
  4 = command not recognized
*/
{
 char *cmd= NULL,line[SfileadrL],*adr,*username= NULL;
 int ret,idx,no_files= 0,yes= 0,count,i;

 if(((flag&4) || Asagent_get_service_mode(app->agent,0)==1) && !(flag&8)) {
   /* client-mode: there are no non-remote commands in ImvapP */
   ret= Asagent_interpreter(app->agent,4|(flag&3));
   return(ret);
 }

 app->agent->input_fd= Asagent_get_event_connection(app->agent,0);
 app->recursion_level++;
 if(app->recursion_level>app->recursion_max){
   Asagent_stderr(app->agent,
            "Command recursion too deep, probably endless -read_from loop.",1);
   ret= 0;
   goto ex;
 }

 ret= Imvapp_determine_current(app,&adr,&idx,&count,0);
 if(ret<=0)
   goto ex;
 else if(ret==2)
   no_files= 1;
 ret= Asagent_get_event_cmd(app->agent,&cmd,0);
 if(ret<=0)
   goto ex;
 if(strncmp(cmd,"--imvapp",8)==0) {
   /* remove own module prefix if there */
   for(i=8;cmd[i]!=0;i++)
     cmd[i-8]= cmd[i];
   cmd[i]= 0;
 }

 ret= 0;
 if(!(flag&1)) {
   if(flag&2) {
     sprintf(line,"%ld> ",Asagent_get_event_connection(app->agent,0));
   } else {
     ret= Asagent_get_current_user(app->agent,&username,1);
     if(ret<=0)
       goto ex;
     sprintf(line,"%s@%ld> ",username,
                             Asagent_get_event_connection(app->agent,0));
   }
   Asagent_stdout(app->agent,line,0);
   Asagent_stdout(app->agent,cmd,1);
 }
 if(cmd[0]==0||cmd[0]=='#'||cmd[0]=='!') {
   ;
 } else if(strncmp(cmd,"-help",5)==0){
   ret= Imvapp_cmd_help(app,strcmp(cmd,"-help_short")==0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-mapper_db",10)==0 && cmd[10]!=0) {
   ret= Asagent_get_permission(app->agent,"service","mapper_db",0);
   if(ret<=0)
     goto ex;
   if(Sregex_string(&(app->mapper_db),cmd+11,0)<=0)
     {ret= -1; goto ex;}
   ret= Imvguard_set_mapper_db(app->guard,app->mapper_db,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-mapper_cmd",11)==0 && cmd[11]!=0) {
   ret= Imvapp_cmd_mapper_cmd(app,cmd+12,0);
   if(ret<=0)
     goto ex;
 } else {
   if(app->mover->event==NULL){
     ret= Asevent_new(&(app->mover->event),"",-1,"",0);
     if(ret<=0)
       goto ex;
   }
   ret= Asevent_copy(app->agent->event,app->mover->event,0);
   if(ret<=0)
     goto ex;
   ret=  Imvmover_interpreter(app->mover,1|2);
   if(ret<=0 || ret==3)
     goto ex;
   if(ret==4) {
     ret= Asagent_interpreter(app->agent,1|2|(flag&(4|8)));
     if(ret<=0 || ret==3)
       goto ex;
   }
   if(ret==4) {
     sprintf(line,"--- command not recognized : '%s'",cmd);
     Asagent_stderr(app->agent,line,1);
     goto ex;
   }
 }
 if(ret==0) ret= 1;
ex:;
 Sregex_string(&cmd,NULL,0);
 Sregex_string(&username,NULL,0);
 app->recursion_level--;
 return(ret);
}


int Imvapp_loop(app,flag)
struct ImvapP *app;
int flag;
/*
 bit0= reset file item index
*/
/*
 return:
 <=0 : error
 1   : ok
 3   : end app service
*/
{
 int ret,idx,count,no_files,not_shown,fstep,show_in_any_case= 0,prompted= 0,ok;
 char *adrpt;

 if(app==NULL)
   return(0);
 if(flag&1)
   Imvmover_reset_cursor(app->mover,0);

 ret= Asagent_loop(app->agent,0);
 return(ret);
}


int Imvapp_read_options(app,argc,argv,ard,flag)
struct ImvapP *app;
int argc;
char **argv;
int *ard;
int flag;
/*
 return:
 <=0 : error
   1 : ok
   3 : end app service
*/
{
 int ret,zero=0;
 char cmd[SfileadrL*2],*cmdpt;

 cmdpt= cmd;

 if(app==NULL)
   return(0);
 for(;(*ard)<argc;(*ard)++){
   if(strlen(argv[*ard])>SfileadrL)
 continue;
   if(argv[*ard][0]=='-')
     strcpy(cmd,argv[*ard]);
   else
     sprintf(cmd,"-addfile:%s",argv[*ard]);
   /* execute cmd as faked option */
   zero= 0;
   ret= Asagent_read_options(app->agent,1,&cmdpt,&zero,0);
   if(ret<=0)
     goto ex;
   if(ret==3)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Imvapp_rc(app,flag)
struct ImvapP *app;
int flag;
{
 int ret;
 char name[SfileadrL],*cmd= NULL;

 if(Sfile_home_adr(".simv_rc",name,0)<=0)
   return(2);
 if(Sregex_string(&cmd,"-readfrom:",0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&cmd,name,1)<=0)
   {ret= -1; goto ex;}
 ret= Asagent_dummy_event(app->agent,cmd,-1,"internal_owner",0);
 if(ret<=0)
   goto ex;
 ret= Asagent_cmd_readfrom(app->agent,name,-1,"internal_owner",0);
ex:
 Sregex_string(&cmd,NULL,0);
 return(ret);
}


