
/*
 cc -g -c sconn.c
*/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

/* in case the OS needs distinction of files and sockets . Linux does not. */
#define closesocket close

#include "smem.h"
#include "sfile.h"
#include "stcp.h"
#include "stunnel.h"
#include "saccess.h"

#include "sconn.h"



/* ------------------------------ SconnfD --------------------------- */


int Sconnfd_new(cfd,type,filename,fd,flag)
struct SconnfD **cfd;
int type;
char *filename;
int fd;
int flag;
{
 struct SconnfD *c;

 *cfd= c= TSOB_FELD(struct SconnfD,1);
 if(c==NULL)
   return(-1);
 c->references= 1;
 c->type= type;
 c->fd= fd;
 c->borrowed_from= NULL;
 c->filename= NULL;
 c->buffer= NULL;
 c->bufflimit= Sconnfd_writebufmaX;
 c->buffsize= 0;
 c->buffcount= 0;
 c->min_writerate=  Sconnfd_min_writeratE;
 c->eof_reached= 0;

 if(Sregex_string(&(c->filename),filename,0)<=0)
   return(-1);

 return(1);
}


int Sconnfd_destroy(cfd,flag)
struct SconnfD **cfd;
int flag;
{
 struct SconnfD *c;

 c= *cfd;
 if(c==NULL)
   return(0);

 if(c->references>1) {
   c->references--;
   *cfd= NULL;
   return(2);
 }

 if(c->borrowed_from==NULL) 
   Sconnfd_close(c,0);
 else
   Sconnfd_destroy(&(c->borrowed_from),0);
 Sregex_string(&(c->filename),NULL,0);
 if(c->buffer!=NULL)
   Smem_freE((char *) (c->buffer));
 Smem_freE((char *) *cfd);
 *cfd= NULL;
 return(1);
}


int Sconnfd_clone(cfd,original,flag)
struct SconnfD **cfd;
struct SconnfD *original;
int flag;
{
 int ret;
 struct SconnfD *c;


 if(original==NULL)
   return(0);
 ret= Sconnfd_new(cfd,5,"",-1,0);
 if(ret<=0)
   return(ret);

 c= *cfd;
 c->type= original->type;
 c->fd= original->fd;
 Sregex_string(&(c->filename),original->filename,0);
 c->eof_reached= original->eof_reached;

 c->borrowed_from= original;
 original->references++;
 return(1);
}

int Sconnfd_close(cfd,flag)
struct SconnfD *cfd;
int flag;
{

 if(cfd->type==0) {
   if(cfd->fd>=0) 
     close(cfd->fd);
   cfd->fd= -1;
 } else if(cfd->type==1) {
   close(cfd->fd);
   if(cfd->filename!=NULL)
     unlink(cfd->filename);
 } else if(cfd->type==2) {
   Stcpfd_close(cfd->fd,0);
 } else if(cfd->type==3) {
   Stcpfd_close(cfd->fd,0);
 } else if(cfd->type==4) {
   ;
 } else if(cfd->type==5) {
   ;
 }
 return(1);
}


int Sconnfd_is_valid(cfd,flag)
struct SconnfD *cfd;
int flag;
{
 if(cfd==NULL)
   return(0);
 if(cfd->fd<0)
   return(0);
 if(cfd->type==4)
   return(0);
 if(cfd->eof_reached)
   return(0);
 return(1);
}


int Sconnfd_read(cfd,buf,bufsize,count,flag)
struct SconnfD *cfd;
char *buf;
int bufsize;
int *count;
int flag;
/*
flag:
 bit0= no need to check wether fd is ready

return:
 <0 = error
  0 = eof
  1 = data read
  2 = no data available (would block)
*/
{
 int ret= 0;

 if(!Sconnfd_is_valid(cfd,0))
   return(0);
 if(!(flag&1))
   if(!Sfile_fd_is_ready(cfd->fd,0,1)>0)
     return(2);
 if(cfd->type==3) 
   ret= Stcpfd_read(cfd->fd,buf,bufsize,count,0);
 else if(cfd->type==0 || cfd->type==1 || cfd->type==5)
   ret= Sfile_read(cfd->fd,buf,bufsize,count,0);
 if(ret<=0)
   return(ret);
 if(count==0)
   return(2);
 return(1);
}


int Sconnfd_readchar(cfd,reply,flag)
struct SconnfD *cfd;
char *reply;
int flag;
/* >>> should be implemented via Sconn_read() now
flag:
 bit0= no need to check wether fd is ready

return:
 <0 = error
  0 = eof
  1 = character read
  2 = no character available (would block)
*/
{
 int ret= 0,count;

 if(!Sconnfd_is_valid(cfd,0))
   return(0);
 if(!(flag&1))
   if(!Sfile_fd_is_ready(cfd->fd,0,1)>0)
     return(2);
 ret= 0;
 if(cfd->type==3) 
   ret= Stcpfd_read(cfd->fd,reply,1,&count,0);
 else if(cfd->type==0 || cfd->type==1 || cfd->type==5)
   ret= Sfile_read(cfd->fd,reply,1,&count,0);
 if(ret<=0)
   return(ret);
 if(count==0)
   return(2);
 return(1);
}


int Sconnfd_readline(cfd,line,flag)
struct SconnfD *cfd;
char **line;
int flag;
/*
 return: <=0 error or EOF with empty line
         1   line complete (possibly completed by EOF)
         2   line not completed yet
*/
{
 int ret,count;
 char gotstr[2];

 ret= Sconnfd_copy_to_buffer(cfd,"",0,0); /* to get it initialized at start */
 if(ret<=0)
   return(ret);

 while(Sfile_fd_is_ready(cfd->fd,0,1)>0){
   count= 0;
   ret= Sconnfd_readchar(cfd,gotstr,1);
   if(ret<0)
     return(ret);
   if(ret==0){
     cfd->eof_reached= 1;
     if(cfd->buffcount>0) {
       gotstr[0]= '\n';
       count= 1;
     } else
       return(0);
   } else if(ret==1)
     count= 1;
   if(count==0)
 break;
   if(cfd->buffcount+2>=cfd->buffsize) {
     /* >>> enlarge buffer */;
   }
   if(gotstr[0]=='\n' || gotstr[0]==0 
      || cfd->buffcount+2>=cfd->buffsize) {
     if(cfd->buffcount>0)
       if(cfd->buffer[cfd->buffcount-1]=='\r')
         cfd->buffcount--;
     cfd->buffer[cfd->buffcount]= 0;
     if(Sregex_string(line,cfd->buffer,0)<=0)
       return(-1);
     cfd->buffcount= 0;
     cfd->buffer[cfd->buffcount]= 0;
     return(1);
   }
   cfd->buffer[cfd->buffcount]= gotstr[0];
   cfd->buffcount++;
 }
 cfd->buffer[cfd->buffcount]= 0;
 return(2);
}


int Sconnfd_write(cfd,data,size,flag)
struct SconnfD *cfd;
char *data;
int size;
int flag;
/*
flag:
 bit0=  check wether fd is ready , if not ready: do not buffer and return 0
 bit2= emulate blocking i/o , do not return 2
return:
 <0 = error
  0 = write buffer overflow
  1 = character read
  2 = not sent completely (would block), buffered for next write attempt
*/
{
 int ret,i,timeout=0;
 double Sfile_microtime(),start_time,end_time,time_to_wait;

 if(cfd->min_writerate>0) {
   start_time= Sfile_microtime(0);
   end_time= start_time+size/cfd->min_writerate;
 } else
   flag|= 4;

try_again:;
 if((flag&1))
   timeout= 1000000;
 /* try to send deferred data first */
 if(cfd->buffcount>0) {
   if(!Sfile_fd_is_ready(cfd->fd,timeout,2)) {
     if(flag&1)
       return(0);
   } else {
     ret= write(cfd->fd,cfd->buffer,cfd->buffcount);
     if(ret<0) {
       if(errno!=EAGAIN)
         goto ex;
     } else if(ret>0 && ret<cfd->buffcount) {
       for(i=ret;i<cfd->buffcount;i++)
         cfd->buffer[i-ret]= cfd->buffer[i];
       cfd->buffcount-= ret;
     } else if(ret>0 && ret>=cfd->buffcount) {
       cfd->buffcount= 0;
     }
   }
 }

 if((flag&1))
   if(!Sfile_fd_is_ready(cfd->fd,1000000,2)>0)
     return(0);

 if(size>0) {
   if(cfd->buffcount>0)               /* do not write if buffer is not empty */
     {ret= 0; goto buffer_remainder;}
   ret= write(cfd->fd,data,size);
 } else
   ret= 0;
 if(ret<0) {
   if(errno==EAGAIN)
     {ret= 0; goto buffer_remainder;}
   goto ex;
 }
 if(ret<size) {
buffer_remainder:;
   if(cfd->buffcount>0 && cfd->buffcount+size>cfd->bufflimit)
     {ret= 0; goto ex;}
   ret= Sregex_concat_bin(&(cfd->buffer),&(cfd->buffcount),data+ret,size-ret,0);
   if(ret<=0)
     {ret= -1; goto ex;}
 }
 if(cfd->buffcount>0)
   ret= 2;
 else
   ret= 1;
ex:;
 if(ret==2) {
   if((flag&4))
     {size= 0; goto try_again;}
   if(cfd->min_writerate>0) if((time_to_wait= end_time-Sfile_microtime(0))>0) {
     if(time_to_wait>0.001)
       Sfile_microsleep(1000,0);
     else
       Sfile_microsleep((int) (time_to_wait*1.0e6));
     {size= 0; goto try_again;}
   }
 }
 return(ret);
}


int Sconnfd_get_typestring(cfd,typestring,flag)
struct SconnfD *cfd;
char **typestring;
int flag;
/*
 bit0= set typestring by Sregex_string rather than by simple assignment
*/
{
 static char types[][16]= {Sconnfd_typelisT};

 if(cfd==NULL) {
   if(flag&1) {
     if(Sregex_string(typestring,types[0],0)<=0)
       return(-1);
   } else   
     *typestring= types[0];
   return(0);
 }
 if(flag&1) {
   if(Sregex_string(typestring,types[cfd->type],0)<=0)
     return(-1);;
 } else
   *typestring= types[cfd->type]; 
 return(1);
}


int Sconnfd_copy_to_buffer(cfd,data,datacount,flag)
struct SconnfD *cfd;
char *data;
int datacount;
int flag;
{
 int i;

 if(cfd==NULL)
   return(0);
 if(cfd->buffer==NULL) { /* initialize line buffer */
   cfd->buffer= TSOB_FELD(char,Sconnfd_linesizE);
   if(cfd->buffer==NULL)
     return(-1);
   cfd->buffsize= Sconnfd_linesizE;
   cfd->buffcount= 0;
 }
 if(datacount+cfd->buffcount>=cfd->buffsize)
   return(0);
 for(i=0;i<datacount;i++)
   cfd->buffer[cfd->buffcount++]= data[i];
 return(1);
}



/* ------------------------------- SconN --------------------------- */


int Sconn_new(conn,id,access,flag)
struct SconN **conn;
char *id;
struct SaccesS *access;
int flag;
{
 struct SconN *c;

 *conn= c= TSOB_FELD(struct SconN,1);
 if(c==NULL)
   return(-1);

 c->id= NULL;
 c->references= 0;
 c->access= access;
 c->tunnel_required= 0;
 c->tunnel_allowed= 0;
 c->tunnel_expect_zero= 0;
 c->tunnel_read_buf= NULL;
 c->tunnel_read_pt=  c->tunnel_read_buf;
 c->tunnel_read_count= 0;
 c->tunnel= NULL;
 c->tunnel_failed= 0;
 c->in_cfd= NULL;
 c->out_cfd= NULL;

 if(Sregex_string(&(c->id),id,0)<=0)
   goto ex;
 if(Sregex_string(&(c->tunnel_read_buf),"-handshake:and some more bytes",0)<=0)
   goto ex;
 c->tunnel_read_pt=  c->tunnel_read_buf;
 
 return(1);
ex:
 Sconn_destroy(conn,0);
 return(-1);
}


int Sconn_destroy(conn,flag)
struct SconN **conn;
int flag;
{
 struct SconN *c;

 c= *conn;
 if(c==NULL)
   return(0);

 if(c->references>1) {
   c->references--;
   *conn= NULL;
   return(2);
 }
 Stunnel_destroy(&(c->tunnel),0);
 Sconnfd_destroy(&(c->in_cfd),0);
 Sconnfd_destroy(&(c->out_cfd),0);
 Sregex_string(&(c->id),NULL,0);
 Sregex_string(&(c->tunnel_read_buf),NULL,0);
 Smem_freE((char *) c);
 *conn= NULL;
 return(1);
}


int Sconn_is_valid(conn,flag)
struct SconN *conn;
int flag;
/*
 bit0= test outbound fd
*/
{
 struct SconnfD *cfd;
 int ret;

 if(conn==NULL)
   return(0);
 if(conn->tunnel_failed)
   return(0);
 if(flag&1)
   cfd= conn->out_cfd;
 else
   cfd= conn->in_cfd;
 ret= Sconnfd_is_valid(cfd,0);
 return(ret);
}


int Sconn_has_bytes(conn,flag)
struct SconN *conn;
int flag;
{
 int count= 0;

 if(conn->tunnel_read_count>0)
   count= conn->tunnel_read_count-(conn->tunnel_read_pt-conn->tunnel_read_buf);
 return(count);
}


int Sconn_get_id(conn,id,flag)
struct SconN *conn;
char **id;
int flag;
{
 int ret;

 if(conn==NULL)
   return(0);
 if(conn->id==NULL)
   ret= Sregex_string(id,"",0);
 else
   ret= Sregex_string(id,conn->id,0);
 return(ret);
}


int Sconn_get_filename(conn,id,flag)
struct SconN *conn;
char **id;
int flag;
/*
 bit0= get from outbound cfd
*/
{
 int ret;
 struct SconnfD *cfd;

 if(conn==NULL)
   return(0);
 if(flag&1)
   cfd= conn->out_cfd;
 else
   cfd= conn->in_cfd;
 if(cfd==NULL) 
   ret= Sregex_string(id,"",0);
 else if(cfd->filename==NULL)
   ret= Sregex_string(id,"",0);
 else
   ret= Sregex_string(id,cfd->filename,0);
 return(ret);
}


int Sconn_get_typestring(conn,typestring,flag)
struct SconN *conn;
char **typestring;
int flag;
/*
 bit0= get outbound type
 bit1= set typestring by Sregex_string rather than by simple assignment
*/
{ 
 int ret;
 struct SconnfD *cfd;

 if(conn==NULL)
   return(-1);
 if(flag&1)
   cfd= conn->out_cfd;
 else
   cfd= conn->in_cfd;
 ret= Sconnfd_get_typestring(cfd,typestring,(flag&2)>>1);
 return(ret);
}


int Sconn_get_fd(conn,flag)
struct SconN *conn;
int flag;
/*
 bit0= get outbound fd
 bit1= get fd despite it is not valid 
 bit2= with bit0: get inbound fd if suitable and outbound fd invalid 
*/
{
 struct SconnfD *cfd;

 if(conn==NULL)
   return(-1);
 if(flag&1)
   cfd= conn->out_cfd;
 else
   cfd= conn->in_cfd;
 if(!(flag&2))
   if(!Sconnfd_is_valid(cfd,0)) {
     if((flag&5)!=5)
       return(-1);
     if(!Sconnfd_is_valid(conn->in_cfd,0))
       return(-1);
     if(conn->in_cfd->type!=3)
       return(-1);
     cfd= conn->in_cfd;
   }
 if(cfd!=NULL)
   return(cfd->fd);
 return(-1);
}


int Sconn_get_filetype(conn,flag)
struct SconN *conn;
int flag;
/*
 bit0= get outbound
 bit1= get despite it is not valid 
 bit2= with bit0: get inbound if outbound invalid 
*/
{
 struct SconnfD *cfd;

 if(conn==NULL)
   return(-1);
 if(flag&1)
   cfd= conn->out_cfd;
 else
   cfd= conn->in_cfd;
 if(!(flag&2))
   if(!Sconnfd_is_valid(cfd,0)) {
     if(!(flag&(1|4)))
       return(-1);
     if(!Sconnfd_is_valid(conn->in_cfd,0))
       return(-1);
     cfd= conn->in_cfd;
   }
 if(cfd!=NULL)
   return(cfd->type);
 return(-1);
}


int Sconn_get_username(conn,result,flag)
struct SconN *conn;
char **result;
int flag;
{
 int ret;

 if(conn->tunnel==NULL){
   ret= Sregex_string(result,"external_anonymous",0);
   return(ret);
 }
 ret= Stunnel_get_username(conn->tunnel,result,0);
 return(ret);
}


int Sconn_attach_fd(conn,type,filename,fd,flag)
struct SconN *conn;
int type;
char *filename;
int fd;
int flag;
/*
 bit0= attach to out_cfd
*/
{
 int ret;
 char *typestring,*value,*name= NULL;
 struct SconnfD *cfd= NULL,**cfdpt,**other_cfdpt;

 if(conn==NULL)
   return(0);
 if(flag&1){
   cfdpt= &(conn->out_cfd);
   other_cfdpt= &(conn->in_cfd);
 } else {
   cfdpt= &(conn->in_cfd);
   other_cfdpt= &(conn->out_cfd);
 }
 if((*other_cfdpt)!=NULL)
   if((*other_cfdpt)->fd==fd) {
     ret= Sconnfd_clone(&cfd,*other_cfdpt,0);
     if(ret<=0)
       goto ex;
   }
 if(cfd==NULL) {
   ret= Sconnfd_new(&cfd,type,filename,fd,0);
   if(ret<=0)
     goto ex;
 }
 Sconnfd_destroy(cfdpt,0);
 if(!(flag&1)) {
   Sconnfd_get_typestring(cfd,&typestring,0);
   ret= Saccess_find_value(conn->access,"connection",typestring,"tunnel",
                           &value,0);
   if(ret==0) {
     if(strcmp(typestring,"file")==0 && cfd->filename!=NULL) {
       if(Sregex_string(&name,"<",0)<0)
         {ret= -1; goto ex;}
       if(Sregex_string(&name,cfd->filename,1)<=0)
         {ret= -1; goto ex;}
       ret= Saccess_find_value(conn->access,"connection",name,"tunnel",
                               &value,0);
     }
   }
   if(ret<0)
     goto ex;
   if(ret>0) {
     if(strcmp(value,"require")==0)
       conn->tunnel_required= 1;
     if(strcmp(value,"allow")==0)
       conn->tunnel_allowed= 1;
   }
 }
 (*cfdpt)= cfd;
 ret= 1;
ex:;
 Sregex_string(&name,NULL,0);
 return(ret);
}


int Sconn_detach_cfd(conn,flag)
struct SconN *conn;
int flag;
/*
 bit0= detach out_cfd
*/
{
 int ret;
 struct SconnfD **cfdpt;

 if(conn==NULL)
   return(0);
 if(flag&1)
   cfdpt= &(conn->out_cfd);
 else
   cfdpt= &(conn->in_cfd);
 Sconnfd_destroy(cfdpt,0);
 return(1);
}
 

int Sconn_refer(conn,referer,flag)
struct SconN *conn;
struct SconN **referer;
int flag;
{
 *referer= conn;
 conn->references++;
 return(1);
}


int Sconn_unrefer(conn,flag)
struct SconN **conn;
int flag;
{
 Sconn_destroy(conn,0); /* decrements reference and destroys if 0 */
 *conn= NULL;
 return(1);
}


int Sconn_queue_tunnelerr(conn,tunnel,flag)
struct SconN *conn;
struct StunneL *tunnel;
int flag;
{
 int ret=2;
 char fd_text[81];

 if(conn==NULL)
   return(0);
 if(conn->access==NULL || tunnel==NULL)
   return(0);
 if(tunnel->errmsg[0]!=0) {
   sprintf(fd_text,"%ld tunnel",Sconn_get_fd(conn,0));
   ret= Saccess_append(conn->access,"message","stderr",fd_text,
                       tunnel->errmsg,0);
   if(ret<=0)
     return(ret);
   tunnel->errmsg[0]= 0;
 }
 return(ret);
}


int Sconn_try_tunnel(conn,flag)
struct SconN *conn;
int flag;
/*
 return:
  <0 = error
   0 = handshake failed
   1 = handshake succeeded
   2 = no tunnel established
*/
{
 int ret,handshake_len= 0,buf_count= 0,rb_len= 0,i,write_len,sc_len;
 char *username= NULL,*keydir= NULL,*typestring= NULL,*handshake= NULL;
 char *endpt,*rb_buf= NULL;
 char buf[Stunnel_max_tokenleN];
 struct StunneL *tunnel= NULL;
 
 /* lookup access wether a user is set */
 ret= Sconn_get_typestring(conn,&typestring,1);
 if(ret<=0)
   goto ex;
 ret= Saccess_find_value(conn->access,"connection",typestring,"user",
                         &username,0);
 if(ret<0)
   goto ex;
 if(username==NULL) 
   {ret= 2; goto ex;}
 if(username[0]==0) 
   {ret= 2; goto ex;}
 ret= Saccess_find_value(conn->access,"connection",typestring,"keydirectory",
                         &keydir,0);
 if(ret<=0)
   goto ex;

 /* init and accept handshake */;
 ret= Stunnel_new(&tunnel,username,Sconn_get_fd(conn,1),keydir,0);
 if(ret<=0)
   goto ex;
 ret= Stunnel_handshake_init(tunnel,&handshake,&handshake_len,"-",0);
 if(ret<=0)
   goto ex;
 write_len= handshake_len;
 for(i=0;i<Sconn_handshake_timeouT;i++) {
   ret= Sconn_write(conn,handshake,write_len,2);
   if(ret<=0)
     goto ex;
   if(ret!=2)
 break;
   Sfile_microsleep(1000000,0);
   write_len= 0;               /* data is buffered after first Sconn_write() */
 }
 if(i>=Sconn_handshake_timeouT)
   {ret= -1; goto ex;}                                            /* timeout */
 while(1) { 
   if(!Sfile_fd_is_ready(conn->in_cfd->fd,Sconn_handshake_timeouT*1000000,1))
     {ret= 0; goto ex;}
   ret= Sconn_read(conn,buf,sizeof(buf),&buf_count,2|1);
   if(ret!=1)
     goto ex;
   ret= Stunnel_handshake_accept(tunnel,buf,buf_count,&endpt,0);
   if(ret<=0)
     goto ex;
   if(ret==1)
 break;
 }  

 if(endpt!=NULL)
   if(endpt-buf<buf_count) {
     /* save remaining data as beginning of tunnel_read_buf */
     if(Sregex_concat_bin(&rb_buf,&rb_len,endpt,buf_count-(endpt-buf),0)<=0)
       {ret= -1; goto ex;}
     sc_len= Sconn_has_bytes(conn,0);
     if(sc_len>0)
       if(Sregex_concat_bin(&rb_buf,&rb_len,conn->tunnel_read_buf,sc_len,0)<=0)
         {ret= -1; goto ex;}
     if(conn->tunnel_read_buf!=NULL)
       Smem_freE(conn->tunnel_read_buf);
     conn->tunnel_read_buf= rb_buf;
     conn->tunnel_read_count= rb_len;
   }
 Stunnel_destroy(&(conn->tunnel),0);
 conn->tunnel= tunnel;
 conn->tunnel_allowed= conn->tunnel_required= 0;
 ret= 1;
ex:;
 if(ret<=0) {
   Sconn_queue_tunnelerr(conn,tunnel,0);
   Stunnel_destroy(&tunnel,0);
 }
 Sregex_string(&handshake,NULL,0);
 return(ret);
}


int Sconn_expect_tunnel(conn,flag)
struct SconN *conn;
int flag;
/* 
 return:
  <=0 : error (e.g. handshake failed)
    1 : tunnel installed
    2 : first line not complete yet
    3 : optional handshake did not occur, data stored in buf
*/
{
 int ret,reply_len,type,fd,i,write_len;
 char gotchar[2],*line= NULL,errmsg[2*SfileadrL],*typestring,*keydir;
 char *backchannel= NULL,fd_text[81];
 static char expected[]= {"-handshake:"};
 struct StunneL *tunnel;

 /* general error message */
 strcpy(errmsg,"Shielded connection required. Next time use -handshake: .");

 gotchar[1]= 0;
 if(conn->tunnel_expect_zero) {
   conn->tunnel_expect_zero= 0;
   ret= Sconnfd_readchar(conn->in_cfd,gotchar,0);
   if(ret<=0)
     goto ex;
   if(gotchar[0]!=0) {
     conn->tunnel_read_buf[0]= gotchar[0];
     conn->tunnel_read_count= 1;
     goto refuse; 
   } else if(conn->tunnel_required==0)
     conn->tunnel_required= 1; 
 }
 if(conn->tunnel_read_count<strlen(expected)) {
   while(conn->tunnel_read_count<strlen(expected)) {
     ret= Sconnfd_readchar(conn->in_cfd,
                         &(conn->tunnel_read_buf[conn->tunnel_read_count]),0);
     if(ret!=1)
       goto ex;
     conn->tunnel_read_count++;
     if(conn->tunnel_read_buf[conn->tunnel_read_count-1]
        != expected[conn->tunnel_read_count-1])
       goto refuse;
   }
   conn->tunnel_read_buf[conn->tunnel_read_count]= 0;
   if(strcmp(conn->tunnel_read_buf,expected)!=0)
     goto refuse;
   ret= Sconnfd_copy_to_buffer(conn->in_cfd,conn->tunnel_read_buf,
                               conn->tunnel_read_count,0);
   if(ret<=0)
     goto ex;  
 }
 /* read remainder of line */ 
 /* from here on, no tunnel_allow behavior:
    the handshake must be correct or the connection has to fail */
 ret= Sconnfd_readline(conn->in_cfd,&line,0);
 if(ret!=1)
   goto ex;
 if(strlen(line)>SfileadrL)                   /* prevents overflow of errmsg */
   goto failed;

 /* negociate tunnel */
 Sconnfd_get_typestring(conn->in_cfd,&typestring,0);
 ret= Saccess_find_keydir(conn->access,typestring,&keydir,0);
 if(ret<=0) {
   strcpy(errmsg,"No key directory defined at server.");
   goto failed;
 }
 ret= Stunnel_handshake_reply(&tunnel,errmsg,&line,&reply_len,&backchannel,
                              conn->in_cfd->fd,keydir,0);
 if(ret<=0)
   goto failed;
 conn->tunnel= tunnel;

 /* >>> connect out_cfd to backchannel */
 if(!Sconn_is_valid(conn,1)) {
   if(strcmp(backchannel,"-")==0 || backchannel[0]==0) {
     /* clone in_cfd */
     ret= Sconn_attach_fd(conn,5,backchannel,conn->in_cfd->fd,1);
     if(ret<=0)
       goto ex;
   } else {
     /* >>> create outward cfd */  ;
     type= 0;
     fd= open(backchannel,O_WRONLY|O_NONBLOCK);
     if(fd<0) {
       /* >>> error message */;
       ret= 0;
       goto ex;
     }
     ret= Sconn_attach_fd(conn,type,backchannel,fd,1);
     if(ret<=0)
       goto ex;
   }
 }

 /* send reply via back channel (without further encryption)*/
 write_len= reply_len;
 for(i=0;i<Sconn_handshake_timeouT;i++) {
   ret= Sconn_write(conn,line,write_len,2);
   if(ret<=0)
     goto failed;
   if(ret!=2)
 break;
   Sfile_microsleep(1000000,0);
   write_len= 0;               /* data is buffered after first Sconn_write() */
 }
 if(i>=Sconn_handshake_timeouT)
   {ret= -1; goto failed;}                                        /* timeout */
 
 conn->tunnel_read_count= 0;
 conn->tunnel_required= 0;
 conn->tunnel_allowed= 0;
 conn->tunnel_failed= 0;
 ret= 1;
ex:;
 Sregex_string(&backchannel,NULL,0);
 Sregex_string(&line,NULL,0);
 return(ret);
refuse:;
 if(conn->tunnel_required) {
failed:;
   ret=0;
   conn->tunnel_failed= 1;
   if(Sconn_is_valid(conn,1)) {
     Sconn_write(conn,errmsg,strlen(errmsg),2);
     Sconn_write(conn,"\n",1,2);
   }
   sprintf(fd_text,"%ld tunnel",Sconn_get_fd(conn,0));
   Saccess_append(conn->access,"message","stderr",fd_text,errmsg,0);
 } else {
   conn->tunnel_allowed= 0;
   ret=3;
 }
 goto ex;
}


int Sconn_reauthenticate(conn,flag)
struct SconN *conn;
int flag;
{
 int ret,tunneled_len;
 char *tunneled= NULL;

 ret= Stunnel_reauthenticate(conn->tunnel,&tunneled,&tunneled_len,0);
 if(ret<=0) {
   Sconn_queue_tunnelerr(conn,conn->tunnel,0);
   return(0);
 }
 ret= Sconn_write(conn,tunneled,tunneled_len,0);
 return(ret);
}


int Sconn_is_ready(conn,flag)
struct SconN *conn;
int flag;
/*
flag:
 bit0= attach to out_cfd
*/
{
 struct SconnfD *cfd;
 int fd,ret;

 if(flag&1)
   cfd= conn->out_cfd;
 else {
   if(Sconn_has_bytes(conn,0)>0)
     return(1);
   cfd= conn->in_cfd;
 }
 if(cfd==NULL)
   return(0);
 fd= cfd->fd;
 if(fd<0)
   return(0);
 ret= Sfile_fd_is_ready(fd,0,1);
 return(ret); 
}


int Sconn_read(conn,buf,bufsize,count,flag)
struct SconN *conn;
char *buf;
int bufsize;
int *count;
int flag;
/*
flag:
 bit0= no need to check wether fd is ready
 bit1= do not decode by conn->tunnel
 bit2= do not read data from fd (only consume remnant buffer)
return:
 <0 = error
  0 = eof
  1 = character read
  2 = no character available (would block)
*/
{
 int ret,linecount= 0,i,w,tbuf_count,type;
 char *line= NULL;

 if(conn->tunnel_failed)
   return(0); 
 if(flag&2)
   goto do_read;

 /* if in tunnel init phase : negociate initial tunnel handshake */
 if(conn->tunnel_required || conn->tunnel_allowed) {
   ret= Sconn_expect_tunnel(conn,0);
   if(ret<=0)
     return(ret);
   if(ret==3)
     ret= 1;
   else if(ret==2)
     return(ret);
   flag&= ~1;             /* doubtful wether there are still data available */
 }

do_read:;
 /* consume tunnel buffer first */
 if(conn->tunnel_read_count>0) {
consume_buffer:;
   tbuf_count= conn->tunnel_read_count-
               (conn->tunnel_read_pt-conn->tunnel_read_buf);
   if(bufsize>=tbuf_count) {
     for(i=0;i<tbuf_count;i++)
       buf[i]= conn->tunnel_read_pt[i];
     *count= tbuf_count;
     conn->tunnel_read_count= 0;
     conn->tunnel_read_pt= conn->tunnel_read_buf;
   } else {
     for(i=0;i<bufsize;i++)
       buf[i]= conn->tunnel_read_pt[i];
     *count= bufsize;
     conn->tunnel_read_pt+= *count;
   }
   return(1);
 }

 if(flag&4)
   return(2);
 ret= Sconnfd_read(conn->in_cfd,buf,bufsize,count,flag);
 if(ret!=1)
   return(ret);

 /* decode and buffer if tunneled */
 if(conn->tunnel!=NULL && !(flag&2)) {
   conn->tunnel_read_pt= NULL;
   ret= Stunnel_retrieve(conn->tunnel,buf,*count,
                        &(conn->tunnel_read_buf),&(conn->tunnel_read_count),0);
   if(ret<=0)
     {ret= -1; goto ex;}
   if(conn->tunnel_read_count>0) {
     conn->tunnel_read_pt= conn->tunnel_read_buf;
     goto consume_buffer;
   }
   *count= 0;
   ret= 2;
 } else
   ret= 1;
ex:;
 if(ret<0)
   Sconn_queue_tunnelerr(conn,conn->tunnel,0);
 return(ret);
}


int Sconn_readchar(conn,reply,flag)
struct SconN *conn;
char *reply;
int flag;
/*
flag:
 bit0= no need to check wether fd is ready

return:
 <0 = error
  0 = eof
  1 = character read
  2 = no character available (would block)
*/
{
 int ret,count;

 ret= Sconn_read(conn,reply,1,&count,flag);
 return(ret);
}


int Sconn_readline(conn,line,flag)
struct SconN *conn;
char **line;
int flag;
/*
 return: <=0 error or EOF with empty line
         1   line complete (possibly completed by EOF)
         2   line not completed yet
*/
{
 struct SconnfD *cfd;
 int ret,count,dummy_count;
 char gotstr[2];

 cfd= conn->in_cfd;
 if(cfd==NULL)
   return(0);

 ret= Sconnfd_copy_to_buffer(cfd,"",0,0); /* to get it initialized at start */
 if(ret<=0)
   return(ret);

 while(Sfile_fd_is_ready(cfd->fd,0,1)>0 || Sconn_has_bytes(conn,0)>0){
   count= 0;
   ret= Sconn_read(conn,gotstr,1,&dummy_count,1);
   if(ret<0)
     return(ret);
   if(ret==0){
     cfd->eof_reached= 1;
     if(cfd->buffcount>0) {
       gotstr[0]= '\n';
       count= 1;
     } else
       return(0);
   } else if(ret==1)
     count= 1;
   if(count==0)
 break;
   if(cfd->buffcount+2>=cfd->buffsize) {
     /* >>> enlarge buffer */;
   }
   if(gotstr[0]=='\n' || gotstr[0]==0
      || cfd->buffcount+2>=cfd->buffsize) {
     if(cfd->buffcount>0)
       if(cfd->buffer[cfd->buffcount-1]=='\r')
         cfd->buffcount--;
     cfd->buffer[cfd->buffcount]= 0;
     if(Sregex_string(line,cfd->buffer,0)<=0)
       return(-1);
     cfd->buffcount= 0;
     cfd->buffer[cfd->buffcount]= 0;
     return(1);                                    /* line finally complete */
   }
   cfd->buffer[cfd->buffcount]= gotstr[0];
   cfd->buffcount++;
 }
 cfd->buffer[cfd->buffcount]= 0;
 return(2);
}


int Sconn_write(conn,data,size,flag)
struct SconN *conn;
char *data;
int size;
int flag;
/*
 bit0= check wether fd is ready
 bit1= do not encode by conn->tunnel
 bit2= emulate blocking i/o , do not return 2

return:
 <0 = error
  0 = write buffer overflow
  1 = character read
  2 = not sent completely (would block), buffered for next write attempt
*/
{
 int ret,tunneled_len,tosend_len;
 char *tunneled= NULL,*tosend;

 if(conn==NULL)
   return(-1);
 if(!Sconnfd_is_valid(conn->out_cfd,0))
   return(-1);

 if(conn->tunnel!=NULL && !(flag&2)) {
   ret= Stunnel_format(conn->tunnel,&tunneled,&tunneled_len,0,data,size,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   tosend= tunneled;
   tosend_len= tunneled_len;
 } else {
   tosend= data;
   tosend_len= size;
 }
 if(tosend_len<=0 && !(flag&2))
   {ret= 1; goto ex;}
 ret= Sconnfd_write(conn->out_cfd,tosend,tosend_len,flag&(1|4));
ex:;
 if(tunneled!=NULL)
   Smem_freE((char *) tunneled);
 if(ret<=0)
   Sconn_queue_tunnelerr(conn,conn->tunnel,0);
 return(ret);
}


int Sconn_tcpclient_new(conn,servername,portnumber,access,flag)
/* Make a new SconN connected to a TCP/IP server */
struct SconN **conn;
char *servername;
int portnumber;
struct SaccesS *access;
int flag;
{
 int ret,fd= -1,fd_attached= 0;
 char *id= NULL,porttext[81];

 *conn= NULL;

 Sregex_string(&id,servername,0);
 sprintf(porttext,"!%ld",portnumber);
 Sregex_string(&id,porttext,1);

 ret= Stcp_clientfd(&fd,servername,portnumber,0);
 if(ret<=0)
   goto ex;
 ret= Sconn_new(conn,id,access,0);
 if(ret<=0)
   goto ex;
 ret= Sconn_attach_fd(*conn,3,id,fd,0);
 if(ret<=0)
   goto ex;
 fd_attached= 1;
 ret= Sconn_attach_fd(*conn,3,id,fd,1);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 if(ret<=0) {
   Sconn_destroy(conn,0);
   if(fd>=0 && !fd_attached)
     Stcpfd_close(fd,0);
 } else {
   (*conn)->references= 1;
 }
 Sregex_string(&id,NULL,0);
 return(ret);
}
