
/*
  cc -g -c sregex.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#ifdef Sregex_use_smeM

#include "smem.h"

#else /* Sregex_use_smeM */

#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));
#define Smem_malloC malloc
#define Smem_freE free

#endif /* ! Sregex_use_smeM */


#include "sregex.h"


/* --------------------- sregex ---------------------- */


int Sregex_new(re,expr,flag)
struct SregeX **re;
char *expr;
int flag;
{
 struct SregeX *r;
 r= TSOB_FELD(struct SregeX,1);
 if(r==NULL)
   return(-1);
 r->compiled= 0;
 *re= r;
 if(expr[0]!=0){
   if(Sregex_comp(*re,expr,0)<=0)
     return(0); 
 }else{
   if(Sregex_comp(*re,".*",0)<=0)
     return(0);
 }
 return(1);
}


int Sregex_destroy(re,flag)
struct SregeX **re;
int flag;
{
 if(*re==NULL)
   return(0);
 Sregex_free(*re,0);
 Smem_freE((char *) *re);
 *re= NULL;
 return(1);
}


int Sregex_comp(re,expr,flag)
struct SregeX *re;
char *expr;
int flag;
{
 re->compiled= 0;
 strcpy(re->expr,expr);
 if(regcomp(&(re->reg),expr,0)!=0)
   return(0);
 re->compiled= 1;
 return(1);
}


int Sregex_free(re,flag)
struct SregeX *re;
int flag;
{
 if(re->compiled==0)
   return(0);
 regfree(&(re->reg));
 re->compiled= 0;
 return(1);
}


int Sregex_replace(re,expr,flag)
struct SregeX **re;
char *expr;
int flag;
{
 int ret;

 if(*re==NULL){
   ret= Sregex_new(re,expr,0);
 }else{
   Sregex_free(*re,0);
   if(expr[0]!=0)
     ret= Sregex_comp(*re,expr,0);
   else
     ret= 1;
 }
 return(ret);
}


int Sregex_is_match(re,text,flag)
struct SregeX *re;
char *text;
int flag;
{
 if(text[0]==0){
   if(re->expr[0]==0) 
     return(1);
   else
     return(0);
 }
 if(regexec(&(re->reg),text,1,re->match,0)==0) 
   return(1); 
 return(0);
}


int Sregex_is_match_t(expr,text,flag)
char *expr;
char *text;
int flag;
{
 struct SregeX *re;
 int ret;
 
 ret= Sregex_new(&re,expr,0);
 if(ret<=0)
   return(-1);
 ret= Sregex_is_match(re,text,0);
 Sregex_destroy(&re,0);
 return(ret);
}


/* ----------- misc help functions ----------- */


int Sregex_string_to_int(text,reply,flag)
char *text;
int *reply;
int flag;
{
 if(text==NULL)
   return(0);
 if(sscanf(text,"%ld",reply)<=0)
   return(0);
 return(1);
}


int Sregex_complines(nakedline,dressedline,flag)
char *nakedline;
char *dressedline;
int flag;
/*
 bit0= strncmp()==0 is sufficient
*/
{
 int l;
 l= strlen(nakedline);
 if(l==0 || strncmp(nakedline,dressedline,l)==0)
   if((flag&1) || dressedline[l]==0 || dressedline[l]=='\r' 
               || dressedline[l]=='\n')
     return(1);
 return(0);
}


int Sregex_read_text(line,smark,emark,text,textsiz,flag)
char *line;
char *smark,*emark;
char *text;
int textsiz;
int flag;
/*
 bit0= do not clear text before trying to read
*/
{
 char *rpt,*ept,*wpt,*maxwpt;

 if(!(flag&1))
   text[0]= 0;
 if(smark[0]==0)
   rpt= line;
 else
   rpt= strstr(line,smark);
 if(rpt==NULL)
   return(0);
 rpt+= strlen(smark);
 if(emark[0]==0)
   ept= line+strlen(line);
 else
   ept= strstr(rpt,emark);
 if(ept==NULL)
   return(0);
 wpt= text;
 maxwpt= text+textsiz-1;
 while(rpt<ept && wpt<maxwpt)
   *(wpt++)= *(rpt++);
 *wpt= 0;
 return(1);
}


int Sregex_trimline(line,flag)
/*
  removes line endings as well as leading and trailing blanks
*/
char *line;
int flag;
/*
 bit0= do not remove line end (protects trailing blanks if line end is present)
 bit1= do not remove leading blanks
 bit2= do not remove trailing blanks
 bit3= remove surrounding quotation marks (after removing line end)
*/
{
 char *cpt,*wpt;
 int l;

 if(!(flag&1)){
   l= strlen(line);
   if(l>0) if(line[l-1]=='\r') line[--l]= 0;
   if(l>0) if(line[l-1]=='\n') line[--l]= 0;
   if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 }
 if(flag&3){
   l= strlen(line);
   if(l>1) if(line[0]==34 && line[l-1]==34) {
     wpt= line;
     cpt= wpt+1;
     while(*cpt!=0) 
       *(wpt++)= *(cpt++);
     line[l-2]= 0;
   }
 }
 if(!(flag&2)){
   wpt= cpt= line;
   while(*(cpt)!=0) {
     if(!isspace(*cpt))
   break;
     cpt++;
   }
   while(*(cpt)!=0) 
     *(wpt++)= *(cpt++);
   *wpt= 0;  
 }
 if(!(flag&4)){
   l= strlen(line);
   if(l<=0)
     return(1);
   cpt= line+l;
   while(cpt-->=line){
     if(!isspace(*cpt))
   break;
     *(cpt)= 0;
   }
 }
 return(1);
}


int Sregex_concat_bin(head,head_len,tail,tail_len,flag)
char **head;
int *head_len;
char *tail;
int tail_len;
int flag;
{
 char *handle;

 if(*head_len<0 || tail_len<0 || *head_len+tail_len<=0)
   return(0);
 handle= TSOB_FELD(char,(*head_len+tail_len+1));
 if(handle==NULL)
   return(-1);
 handle[*head_len+tail_len]= 0;  /* for convenience with human readable data */
 if(*head_len>0)
   memcpy(handle,*head,*head_len);
 if(tail_len>0)
   memcpy(handle+*head_len,tail,tail_len);
 if(*head!=NULL)
   Smem_freE(*head);
 *head= handle;
 (*head_len)+= tail_len;
 return(1);
}


int Sregex_string_cut(handle,text,len,flag)
char **handle;
char *text;
int len;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int l=0;
 char *old_handle;

 if((flag&1)&&*handle!=NULL)
   l+= strlen(*handle);
 old_handle= *handle;
 if(text!=NULL) {
   l+= len;
   *handle= TSOB_FELD(char,l+1);
   if(*handle==NULL) {
     *handle= old_handle;
     return(0);
   }
   if((flag&1) && old_handle!=NULL)
     strcpy(*handle,old_handle);
   else
     (*handle)[0]= 0;
   strncat(*handle,text,len);
 } else {
   *handle= NULL;
 }
 if(old_handle!=NULL)
   Smem_freE(old_handle);
 return(1);
}


int Sregex_string(handle,text,flag)
char **handle;
char *text;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int ret,l=0;

 if(text!=NULL)
   l= strlen(text);
 
/* #define Sregex_looking_for_contenT 1 */
#ifdef Sregex_looking_for_contenT
 /* a debugging point if a certain text content has to be caught */
 if(text!=NULL)
   if(strcmp(text,"clear")==0)
     ret= 0;
#endif

 ret= Sregex_string_cut(handle,text,l,flag&1);
 return(ret);
}


/* old name of Sregex_string . It was just too long */
int Sregex_set_dyn_string(handle,text,flag)
char **handle;
char *text;
int flag;
{
 int ret;
 ret= Sregex_string(handle,text,flag);
 return(ret);
}


int Sregex_search_mark(data,data_len,mark,mark_len,markpt,flag)
char *data;
int data_len;
char *mark;
int mark_len;
char **markpt;
int flag;
{
 int i;

 if(*markpt==NULL)
   return(0);
 if(*markpt>=mark+mark_len)
   return(1);

 i= 0;
 for(i=0;i<data_len;i++) {
   if(**markpt==data[i]) {
     if(*markpt>=mark+mark_len-1)
       return(1);
     (*markpt)++;
   } else 
     *markpt= mark;
 }
 return(0);
}


int Sregex_closing_bracket(text,bracket,flag)
char *text;
char bracket;
int flag;
/*
 bit0= do only respect the privacy of "..."
*/
{
 int i,j,ret,l,list_l,j_start=0;
 int quote= 0,depth= 0,extendable=1; 
 char *stack= NULL,buf[2];
 static char openers[]={"([{'`\""};
 static char closers[]={")]}'`\""};
 static char nesting[]={"yyynnn"};

 buf[1]= 0;
 list_l= strlen(openers);
 l= strlen(text);
 if(flag&1)
   j_start= list_l-1;

 for(i=0;i<l;i++){
   if(depth>0) {
     if(text[i]==closers[stack[depth-1]-1]) { 
       stack[--depth]= 0;
       if(depth>0)
         extendable= (nesting[stack[depth-1]-1]=='y');
       else
         extendable= 1;
     }
   } else if(text[i]==bracket) {
     ret= i;
     goto ex;
   } else if(extendable) {
     for(j=0;j<list_l;j++) {
       if(text[i]==openers[j]){
         buf[0]= j+1;
         ret= Sregex_string(&stack,buf,1);
         if(ret<0)
           goto ex;
         extendable= (nesting[j]=='y');
         depth++;
     break;
       }
     }
   }
 }
 ret= -1;
ex:
 Sregex_string(&stack,NULL,0);
 return(ret);
}


int Sregex_de_html_char(source,target,flag)
char **source,**target;
int flag;
{
 char *spt;
 
 spt= *source;
 if(*spt=='&') {
   if(*(spt+1)=='l') if(*(spt+2)=='t') if(*(spt+3)==';') {
     *((*target)++)= '<';
     (*source)+= 4;
     return(1);
   }
   if(*(spt+1)=='g') if(*(spt+2)=='t') if(*(spt+3)==';') {
     *((*target)++)= '>';
     (*source)+= 4;
     return(1);
   }
   if(*(spt+1)=='q') if(*(spt+2)=='u') if(*(spt+3)=='o') if(*(spt+4)=='t')
                                                         if(*(spt+5)==';') {
     *((*target)++)= '"';
     (*source)+= 6;
     return(1);
   }
   if(*(spt+1)=='a') if(*(spt+2)=='m') if(*(spt+3)=='p') if(*(spt+4)==';') {
     *((*target)++)= '&';
     (*source)+= 5;
     return(1);
   }
 } else {
   *((*target)++)= *((*source)++);
 }
 return(1);
}


int Sregex_de_html(source,target,flag)
char *source,*target;
int flag;
{
 char *bpt,*rpt,*wpt;
 int ret;

 rpt= source;
 wpt= target;
 while(1) {
   bpt= strchr(rpt,'<');
   if(bpt!=NULL) {
     while(rpt<bpt)
       Sregex_de_html_char(&rpt,&wpt,0);
     ret= Sregex_closing_bracket(bpt,'>',1);
     if(ret<=0)
 break;
     rpt= bpt+ret+1;
   } else {
     while(*rpt!=0)
       Sregex_de_html_char(&rpt,&wpt,0);
 break;
   }  
 }
 *wpt= 0;
 return(1);
}


int Sregex_cut_neck(text,neck,head,body,flag)
char *text;
char neck;
char **head; /* this is a malloced copy */
char **body; /* this is just a pointer within text. */
int flag;
/*
 flag:
   bit0= append to head rather than replace it
   bit1= return(0) if text empty
 return:
   <0 = error
    0 = flag:bit1 && text is empty
    1 = neck found, head and body enclosed
    2 = no neck found, all in head, body points to trailing 0
*/
{
 int ret=0;
 char *cpt;

 if(flag&2)
   if(text[0]==0)
     return(0);
 cpt= strchr(text,neck);
 if(cpt==NULL) {
   if(Sregex_string(head,text,flag&1)<=0)
     {ret= -1; goto ex;}
   (*body)= text+strlen(text);
   return(2);
 }
 if(Sregex_string_cut(head,text,(cpt-text),flag&1)<=0)
   {ret= -1; goto ex;}

 ret= 1;
 (*body)= cpt+1;
ex:;
 return(ret);
}


int Sregex_cut_neck_int(text,neck,head,body,flag)
char *text;
char neck;
int *head;
char **body; /* this is just a pointer within text. */
int flag;
/*
 flag:
    bit8= use %X rather than %ld
 return:
    0 = not a usable number
  others see Sregex_cut_neck()
*/
{
 char *name= NULL;
 int ret,cut_ret;

 ret= Sregex_cut_neck(text,neck,&name,body,0);
 if(ret<=0)
   goto ex;
 cut_ret= ret;
 if(strlen(name)==0)
   {ret= 0; goto ex;}
 if(flag&1)
   ret= sscanf(name,"%X",head);
 else
   ret= sscanf(name,"%ld",head);
 if(ret<=0)
   {ret= 0; goto ex;}
 ret= cut_ret;
ex:;
 Sregex_string(&name,NULL,0);
 return(ret);
}


int Sregex_hex_value(text,value,length,flag)
char *text;
unsigned long *value;
int length;
int flag;
{
 int single,i;

 *value= 0;
 if(length>8)
   return(-1);
 if(length>strlen(text))
   return(0);
 for(i=0;i<length;i++) {
   (*value)<<= 4;
   if(text[i]>='0' && text[i]<='9')
     single= text[i]-'0';
   else if (text[i]>='A' && text[i]<='F')
     single= text[i]-'A'+10;
   else if (text[i]>='a' && text[i]<='f')
     single= text[i]-'a'+10;
   else
     return(0);
   (*value)|= single;
 }
 return(1); 
}


/* crc32.c by Craig Bruce, 24-May-92 */
/* found 25 December 1999 as
   http://nic.funet.fi/pub/cbm/crossplatform/converters/unix/crc32.c
   data source altered from file to string: 1 September 2000
*/
int Sregex_crc32(data,len,crc,flag)
char *data;
int len;
unsigned long *crc;
int flag;
/*
 bit0= start of data polynom
*/
{
#define CRC32_GEN       0x04c11db7  /* CCITT standard */
#define CRC32_MSB       0x80000000
       int c, i, crcOverflow, k;
       register unsigned long crcbin;

       if(flag&1)
        crcbin = 0;
       else
        crcbin = *crc;
       for(k=0;k<len;k++) {
                c= data[k];
                for (i=0x80; i; i >>= 1) {
                        crcOverflow = crcbin & CRC32_MSB;
                        crcbin = (crcbin << 1) | (i&c ? 1 : 0);
                        if (crcOverflow) crcbin ^= CRC32_GEN;
                }
       }
 
 *crc= crcbin;
 return(1);
}


/* ---------------------------------- LstrinG --------------------------- */


int Lstring_new(lstring,text,link,flag)
struct LstrinG **lstring;
char *text;
struct LstrinG *link;
int flag;
/*
 bit0= insert before link rather than after it
*/
{
 int ret;
 struct LstrinG *s;

 *lstring= s= TSOB_FELD(struct LstrinG,1);
 if(s==NULL)
   return(-1);
 s->text= NULL;
 s->next= s->prev= NULL;

 if(Sregex_string(&(s->text),text,0)<=0)
   {ret= -1; goto failed;}
 if(link==NULL) {
   ;
 } else if(flag&1) {
   s->next= link;
   s->prev= link->prev;
   if(link->prev!=NULL) 
     link->prev->next= s;
   link->prev= s;
 } else {
   s->prev= link;
   s->next= link->next;
   if(link->next!=NULL) 
     link->next->prev= s;
   link->next= s;
 }
 return(1);
failed:;
 Lstring_destroy(lstring,0);
 return(-1);
}


int Lstring_destroy(lstring,flag)
struct LstrinG **lstring;
int flag;
/*
 bit0= do not set *lstring to NULL
*/
{
 struct LstrinG *s;

 s= *lstring;
 if(s==NULL)
   return(0);
 if(s->prev!=NULL)
   s->prev->next= s->next;
 if(s->next!=NULL)
   s->next->prev= s->prev;
 Sregex_string(&(s->text),NULL,0);
 Smem_freE((char *) s);
 if(!(flag&1))
   *lstring= NULL;
 return(1);
}


int Lstring_destroy_all(lstring,flag)
struct LstrinG **lstring;
int flag;
{
 struct LstrinG *s,*next;

 if((*lstring)==NULL)
   return(0);
 for(s= *lstring; s->prev!=NULL; s= s->prev);
 for(;s!=NULL;s= next){
   next= s->next;
   Lstring_destroy(&s,0);
 }
 *lstring= NULL;
 return(1);
}


int Lstring_find_text(lstring,text,found,flag)
struct LstrinG *lstring;
char *text;
struct LstrinG **found;
int flag;
{
 struct LstrinG *s;

 if(lstring==NULL)
   return(0);
 for(s= lstring; s->prev!=NULL; s= s->prev);
 for(;s!=NULL;s= s->next){
   if(strcmp(s->text,text)==0) {
     *found= s;
     return(1);
   }
 }
 return(0);
}


int Lstring_is_listed(lstring,text,flag)
struct LstrinG *lstring;
char *text;
int flag;
{
 struct LstrinG *s;
 int ret;

 ret=  Lstring_find_text(lstring,text,&s,0);
 return(ret);
} 


