

/*
  cc -g -c stunnel.c
*/

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>



#ifdef Stunnel_use_smeM

#include "smem.h"

#else /* Stunnel_use_smeM */

#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));
#define Smem_malloC malloc
#define Smem_freE free

#endif /* ! Stunnel_use_smeM */



#include "sfile.h"
#include "sregex.h"
#include "sblowfish.h"

#include "stunnel.h"


/* -------------------------------- StunneL ------------------------------ */


int Stunnel_new(tunnel,user,idno,keydir,flag)
struct StunneL **tunnel;
char *user;
int idno;
char *keydir;
int flag;
/*
 flag:
  bit0= do not make personal fish
 return:
  1 = ok
  0 = non final error
 -1 = final error
 -2 = unknown user
*/
{
 struct StunneL *t;
 char key[Stunnel_keyleN+1];
 int i,ret;
 double m;

 if(strlen(user)>Stunnel_userleN)
  return(-2);
 *tunnel= t= TSOB_FELD(struct StunneL,1);
 if(t==NULL)
   return(-1);
 t->user[0]= 0;
 t->authenticated= 0;
 t->keydir= NULL;
 t->idno= idno;
 t->internal_fish= NULL;
 t->random_fish= NULL;
 t->random_count= 0;
 t->personal_fish= NULL;
 t->session_fish= NULL;
 for(i=0;i<=Stunnel_challeN;i++)
   t->challenge[i]= t->reply[i]= t->in_reply[i]= t->in_challenge[i]= 0;
 t->in_status= 0;
 for(i=0;i<Stunnel_headleN;i++)
   t->in_head[i]= 0;
 t->head_count= 0;
 t->data_len= 0;
 t->data_count= 0;
 t->data_buffer= NULL;
 t->in_seqno= 0;
 t->out_seqno= 0;
 t->errmsg[0]= 0;
 t->next= t->prev= NULL;

 ret= Stunnel_set_user(t,user,0);
 if(ret<=0)
   goto failed;
 if(Sregex_string(&(t->keydir),keydir,0)<=0)
   {ret= -1; goto failed;}
 ret= Sblowfish_new(&(t->internal_fish),
                    Sblowfish_make_key(Stunnel_internal_keY,Stunnel_keyleN,
                                       key,0),
                    Stunnel_keyleN,0);
 if(ret<=0)
   goto failed;

 m= ((double) (getpid()&0x7fffffff))+
    ((double) (((unsigned long) tunnel)&0x7fffffff))*251.0;
 Sblowfish_make_random(t->internal_fish,&m,key,Stunnel_keyleN,0);
 ret= Sblowfish_new(&(t->random_fish),key,Stunnel_keyleN,0);
 if(ret<=0)
   goto failed;

 if(!(flag&1)){
   ret= Stunnel_make_personal(t,0);
   if(ret<0)
     goto failed;
   if(ret==0)
     {ret= -2; goto failed;}
 }
 return(1);
failed:
 Stunnel_destroy(tunnel,0);
 return(ret);
}


int Stunnel_destroy(tunnel,flag)
struct StunneL **tunnel;
int flag;
{
 struct StunneL *t;

 if(*tunnel==NULL)
   return(0);
 t= *tunnel;
 Sregex_string(&(t->keydir),NULL,0);
 Sblowfish_destroy(&(t->internal_fish),0);
 Sblowfish_destroy(&(t->random_fish),0);
 Sblowfish_destroy(&(t->personal_fish),0);
 Sblowfish_destroy(&(t->session_fish),0);
 Smem_freE((char *) *tunnel);
 *tunnel= NULL;
 return(1);
}


int Stunnel_set_user(tunnel,username,flag)
struct StunneL *tunnel;
char *username;
int flag;
{
 char userfile[3*Stunnel_userleN+4];

 if(strlen(username)>=Stunnel_userleN)
   return(0);
 strcpy(userfile,username);
 if(strcmp(username,"internal_owner")==0)
   goto illegal_username;
 Sfile_esc_filename(userfile,sizeof(userfile)-1,1|4);
 if(strcmp(userfile,username)!=0 || strchr(username,':')!=NULL) {
illegal_username:;
   sprintf(tunnel->errmsg,"illegal username '%s'",userfile);
   return(0);
 }
 strcpy(tunnel->user,username);
 return(1);
}


int Stunnel_get_username(tunnel,username,flag)
struct StunneL *tunnel;
char **username;
int flag;
{
 if(Sregex_string(username,tunnel->user,0)<=0)
   return(-1);
 if(*username==NULL)
   return(0);
 return(1);
}


int Stunnel_make_personal(tunnel,flag)
struct StunneL *tunnel;
int flag;
{
 int ret;
 char key[Stunnel_keyleN+1];
 char filename[SfileadrL],userfile[3*Stunnel_userleN+4];

 Sblowfish_destroy(&(tunnel->personal_fish),0);

 strcpy(userfile,tunnel->user);
 Sfile_esc_filename(userfile,sizeof(userfile)-1,1|4);
 if(strcmp(userfile,tunnel->user)!=0 || strchr(userfile,':')!=NULL) {
   sprintf(tunnel->errmsg,"illegal username '%s'",userfile);
   return(0);
 }
 sprintf(filename,"%s/%s.tnl",tunnel->keydir,userfile);
 ret= Sfile_file_to_bin(filename,key,Stunnel_keyleN,0);
 if(ret<=0)
   return(ret);
#ifndef Stunnel_no_fishinG
 ret= Sblowfish_transform(tunnel->internal_fish,key,Stunnel_keyleN,1|2);
#endif /* ! Stunnel_no_fishinG */
 if(ret<=0)
   return(ret);
 ret= Sblowfish_new(&(tunnel->personal_fish),key,Stunnel_keyleN,0);
 if(ret<=0)
   return(-1);
 return(1);
}


int Stunnel_make_challenge(tunnel,flag)
struct StunneL *tunnel;
int flag;
{
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),
                       tunnel->challenge,Stunnel_challeN,0);
 return(1);
}


int Stunnel_make_reply(tunnel,challenge,reply,flag)
struct StunneL *tunnel;
char *challenge,*reply;
int flag;
{
 int j,i,c;
 
 /* just see data as 8-byte integers and increment each of them */
 for(j=0;j<Stunnel_challeN;j+=8) {
   for(i=0;i<8;i++) {
     c= (((unsigned char *) challenge)[j+i]+1)&0xff;
     reply[j+i]= c;
     if(c!=0)
   break;
   }
   for(i++;i<8;i++)
     reply[j+i]= challenge[j+i];
 }
 return(1);
}


int Stunnel_format_token(tunnel,message,messagelen,type,data,len,flag)
struct StunneL *tunnel;
char **message;
int *messagelen;
int type;
char *data;
int len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
*/
{
 int ret,i,l=0,buffersize,result_len,transmit_len;
 unsigned long crc;
 char *buffer,typechars[8],seqnochars[16];

 result_len= Stunnel_headleN+len+8*(type!=0);
 result_len= result_len+7-((result_len+7)%8);
 buffersize= result_len+8;
 buffer= Smem_malloC(buffersize);
 if(flag&1)
   Sblowfish_reset_chain(tunnel->personal_fish,2);
 else
   Sblowfish_reset_chain(tunnel->session_fish,2);

 tunnel->out_seqno++;
 for(i=0;i<4;i++)
   seqnochars[3-i]= ((tunnel->out_seqno)>>(8*i))&0xff;
 /* some random bytes are added to obscure the sequence of numbers */
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),
                       seqnochars+4,8,0);       /* only 4 bytes will be used */
 if(buffer==NULL)
   return(-1);
 Sregex_crc32(seqnochars,8,&crc,1);
 if(type) {
   for(i=0;i<4;i++)
     typechars[3-i]= ((type)>>(8*i))&0xff;
   for(i=4;i<8;i++)
     typechars[i]= 0;
   Sregex_crc32(typechars,8,&crc,0);
 } 
 Sregex_crc32(data,len,&crc,0);

 for(i=0;i<4;i++)
   buffer[l+3-i]= (crc>>(8*i))&0xff;
 l+=4;
 if(type)
   transmit_len= -len;
 else
   transmit_len= len;
 for(i=0;i<4;i++)
   buffer[l+3-i]= (transmit_len>>(8*i))&0xff;
 l+= 4;
 for(i=0;i<8;i++)
   buffer[l+i]= seqnochars[i];
 l+= 8;
 if(type){
   for(i=0;i<8;i++)
     buffer[l+i]= typechars[i];
   l+= 8;
 }
 for(i=0;i<len;i++)
   buffer[l+i]= data[i];
 l+= len;
 for(i=l;i<buffersize;i++)
   buffer[i]= 0;
#ifndef Stunnel_no_fishinG
 if(flag&1)
   Sblowfish_transform(tunnel->personal_fish,buffer,result_len,0);
 else
   Sblowfish_transform(tunnel->session_fish,buffer,result_len,0);
#endif /* ! Stunnel_no_fishinG */

 if(*message!=NULL)
   Smem_freE((char *) *message);
 *message= buffer;

#ifdef Stunnel_no_formaT

 memcpy(buffer,data,len);
 *messagelen= len;
 return(1);

#endif /* Stunnel_no_formaT */

 *messagelen= result_len;
 return(1);
}


int Stunnel_format(tunnel,message,messagelen,type,data,len,flag)
struct StunneL *tunnel;
char **message;
int *messagelen;
int type;
char *data;
int len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
*/
{
 int ret,i,tunneled_len,l;
 char cpt,*tunneled=NULL;
 if(len<=Stunnel_max_tokenleN){
   ret= Stunnel_format_token(tunnel,message,messagelen,type,data,len,flag);
   goto ex;
 }

 if(*message!=NULL)
   Smem_freE((char *) *message);
 *message= NULL;
 *messagelen= 0;
 for(i=0;i<len;i+=Stunnel_max_tokenleN){
   l= len-i;
   if(l>Stunnel_max_tokenleN)
     l= Stunnel_max_tokenleN;
   ret= Stunnel_format_token(tunnel,tunneled,tunneled_len,type,data+i,l,0);
   if(ret<=0)
     return(ret);
   if(Sregex_concat_bin(message,messagelen,tunneled,tunneled_len,0)<=0)
     {ret= -1; goto ex;}
 }
 ret= 1;
ex:;
 if(tunneled!=NULL)
   Smem_freE(tunneled);
 return(ret);
}


int Stunnel_retrieve_token(tunnel,in_message,messagelen,endpt,
                           type,data,len,flag)
struct StunneL *tunnel;
char *in_message;
int messagelen;
char **endpt;
int *type;
char **data;
int *len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
 bit1= do not check sequence number

return:
 <0 = error
  0 = authentication failed
  1 = data valid
  2 = no data available yet
*/
{
 int ret,i,l=0,buffersize,datastart,bufferend;
 unsigned char *buffer,*head,*message;
 unsigned long crc,seqno=0;


 message= in_message;  /* to make me feel better if  endpt == &in_message  */
 datastart= 0;
 *len= 0;
 if(*data!=NULL) 
   Smem_freE(*data);
 *data= NULL;
 *endpt= NULL;

 /* init token */
 if(tunnel->in_status==0) {
   tunnel->head_count= 0;
   tunnel->in_status= 1;
   tunnel->data_len= 0;
   tunnel->data_count= 0;
   if(flag&1)
     Sblowfish_reset_chain(tunnel->personal_fish,1);
   else
     Sblowfish_reset_chain(tunnel->session_fish,1);
 }

 /* assemble and decode head */
 if(tunnel->in_status==1) {
   if(messagelen+tunnel->head_count<Stunnel_headleN) {
     for(i=0;i<messagelen;i++)
       tunnel->in_head[tunnel->head_count+i]= message[i];
     tunnel->head_count+= messagelen;
     return(2);
   }
   for(i=0;i<Stunnel_headleN-tunnel->head_count;i++)
     tunnel->in_head[tunnel->head_count+i]= message[i];
   datastart= i;
   tunnel->head_count= Stunnel_headleN;
#ifndef Stunnel_no_fishinG
   if(flag&1)
     Sblowfish_transform(tunnel->personal_fish,
                         tunnel->in_head,Stunnel_headleN,1);
   else
     Sblowfish_transform(tunnel->session_fish,
                         tunnel->in_head,Stunnel_headleN,1);
#endif /* ! Stunnel_no_fishinG */
   head= tunnel->in_head;
   l= 0;
   tunnel->data_crc= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   l= 4;
   tunnel->data_len= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   tunnel->in_status= 2;
   if(tunnel->data_len<-Stunnel_max_tokenleN
      || tunnel->data_len>Stunnel_max_tokenleN ) {
     sprintf(tunnel->errmsg,
             "--- tunnel %ld : recieved oversized token : %ld > %ld\n",
             tunnel->idno,
             (tunnel->data_len>0?tunnel->data_len:-tunnel->data_len),
             Stunnel_max_tokenleN);
     return(0);
   }
   if(tunnel->data_len<0) {
     tunnel->in_status= 3;
     tunnel->data_len= 8-tunnel->data_len;
   }
   l= 8;
   seqno= (head[l+0]<<24)|(head[l+1]<<16)|(head[l+2]<<8)|head[l+3];
   if(!(flag&2))
     if(tunnel->in_seqno+1!=seqno) {
       sprintf(tunnel->errmsg,
       "--- tunnel %ld : sequence number mismatch : %-8.8X should be %-8.8X\n",
               tunnel->idno,seqno,tunnel->in_seqno+1);
       return(0);
     }
   tunnel->in_seqno= seqno;

   if(tunnel->data_len==0) {
     tunnel->in_status= 0;
     return(1);
   }
   buffersize= tunnel->data_len+7-((tunnel->data_len+7)%8)+8;
   buffer= Smem_malloC(buffersize);
   if(buffer==NULL)
     return(-1);
   if(tunnel->data_buffer!=NULL)
     Smem_freE((char *) tunnel->data_buffer);
   tunnel->data_buffer= buffer;

 } 
 if(datastart>=messagelen)
   return(2);

 /* read payload data */
 buffer= tunnel->data_buffer;
 if(tunnel->in_status==2 || tunnel->in_status==3) {
   *type= 0;

   for(i=datastart;i<messagelen;i++) {
     if(tunnel->data_count>=tunnel->data_len && (tunnel->data_count%8)==0)
   break;
     buffer[tunnel->data_count++]= message[i];
   }
   if(i<messagelen) 
     *endpt= message+i;
   if(!(tunnel->data_count>=tunnel->data_len && (tunnel->data_count%8)==0))
     return(2);
#ifndef Stunnel_no_fishinG
   if(flag&1)
     Sblowfish_transform(tunnel->personal_fish,buffer,tunnel->data_count,1);
   else
     Sblowfish_transform(tunnel->session_fish,buffer,tunnel->data_count,1);
#endif /* ! Stunnel_no_fishinG */

   Sregex_crc32(tunnel->in_head+8,8,&crc,1);
   Sregex_crc32(buffer,tunnel->data_len,&crc,0);
   if(tunnel->data_crc!=crc) {
     sprintf(tunnel->errmsg,
             "--- tunnel %ld : crc mismatch : %-8.8X should be %-8.8X\n",
             tunnel->idno, crc,tunnel->data_crc);
     return(0);
   }

   /* read control type */
   if(tunnel->in_status==3) {
     l= 0;
     *type= (buffer[l+0]<<24)|(buffer[l+1]<<16)|(buffer[l+2]<<8)|buffer[l+3];
     for(i=8;i<tunnel->data_len;i++)
       buffer[i-8]= buffer[i];
     tunnel->data_len-= 8;
     if(tunnel->data_len<0)
       tunnel->data_len= 0;
   }

   /* transfer data buffer from tunnel to reply parameter */
   *data= (char *) tunnel->data_buffer;
   *len= tunnel->data_len;
   tunnel->data_buffer= NULL;
   tunnel->data_len= tunnel->data_count= 0;
   tunnel->in_status= 0;
   if(!(flag&2))
     tunnel->authenticated= 1;
   return(1);
 }

 return(2-(tunnel->in_status==0));
}


int Stunnel_retrieve(tunnel,in_message,messagelen,data,len,flag)
struct StunneL *tunnel;
char *in_message;
int messagelen;
char **data;
int *len;
int flag;
/*
 bit0= use personal_fish rather than session_fish
 bit1= do not check sequence number

return:
 <0 = error
  0 = authentication failed
  1 = data valid
  2 = no data available yet
*/
{
 int ret,untunneled_len=0,type;
 char *untunneled= NULL,*rpt;

 rpt= in_message;
 if(*data!=NULL)
   Smem_freE(*data);
 *data= NULL;
 *len= 0;
 while(rpt!=NULL && messagelen-(rpt-in_message)>0) {
   ret= Stunnel_retrieve_token(tunnel,rpt,messagelen-(rpt-in_message),&rpt,
                         &type,&untunneled,&untunneled_len,flag&3);
   if(ret<=0)
     goto ex;

   if(type!=0) {
     /* >>> handle control messages
        (handshake-reply and session-key are not expected here) */
     if(type==3) {
       /* authentication opportunity message : ignore */;
     }
 continue;
   }
   if(untunneled_len>0)
     if(Sregex_concat_bin(data,len,untunneled,untunneled_len,0)<=0)
       {ret= -1; goto ex;}
 }
 if(*len>0)
   ret= 1;
 else
   ret= 2;
ex:;
 if(untunneled!=NULL)
   Smem_freE(untunneled);
 return(ret);
}


int Stunnel_handshake_init(tunnel,message,message_len,backchannel,flag)
struct StunneL *tunnel;
char **message;
int *message_len;
char *backchannel;
int flag;
{
 int ret,i;
 char line[2*SfileadrL],dibmmfohf[Stunnel_challeN+1];

 Stunnel_make_challenge(tunnel,0);
 sprintf(line,"-handshake:0.0 kl=%ld cl=%ld ml=%ld:%s:",
             Stunnel_keyleN,Stunnel_challeN,Stunnel_max_tokenleN,tunnel->user);
 for(i=0;i<Stunnel_challeN;i++)
   dibmmfohf[i]= tunnel->challenge[i];
#ifndef Stunnel_no_fishinG
 Sblowfish_reset_chain(tunnel->personal_fish,2);
 Sblowfish_transform(tunnel->personal_fish,dibmmfohf,Stunnel_challeN,0);
#endif /* ! Stunnel_no_fishinG */
 for(i=0;i<Stunnel_challeN;i++)
   sprintf(line+strlen(line),"%-2.2X",dibmmfohf[i]&0xff);
 sprintf(line+strlen(line),":%s\n",backchannel);

 if(Sregex_string(message,line,0)<=0)
   return(-1);
 *message_len= strlen(*message);
 return(1);
}


int Stunnel_handshake_reply(tunnel,errmsg,message,message_len,backchannel,
                            idno,keydir,flag)
struct StunneL **tunnel;
char errmsg[SfileadrL*2];            
char **message;
int *message_len;
char **backchannel;
int idno;
char *keydir;
int flag;
{
 int ret,i,c,reply_len= 0,skey_reply_len= 0;
 char line[SfileadrL],challenge[Stunnel_challeN+1],key[Stunnel_keyleN+1];
 char *cpt,*gotstr= NULL,*user= NULL,rev[81],*reply= NULL,*skey_reply= NULL;
 static char general_error[]= 
                       {"--- tunnel %ld : malformed handshake attempt : %s\n"};
 struct StunneL *t=NULL;

 cpt= *message;
 errmsg[0]= 0;

 /* verify protocol */
 ret= Sregex_cut_neck(cpt,':',&gotstr,&cpt,0);
 if(ret!=1)
   {ret= 0; goto ex;}
 if(strcmp(gotstr,"-handshake")!=0)
   {ret= 0; goto ex;}
 ret= Sregex_cut_neck(cpt,':',&gotstr,&cpt,0);
 if(ret!=1)
   {ret= 0; goto ex;}
 /* for now, the protocol range is quite small :) */
 sprintf(rev,"0.0 kl=%ld cl=%ld ml=%ld",
             Stunnel_keyleN,Stunnel_challeN,Stunnel_max_tokenleN);
 if(strcmp(gotstr,rev)!=0)
   {ret= 0; goto ex;}

 /* read user */
 ret= Sregex_cut_neck(cpt,':',&user,&cpt,0);
 if(ret!=1)
   {ret= 0; goto ex;}

 /* read in_challenge */
 if(strlen(cpt)<2*Stunnel_challeN)
   {ret= 0; goto ex;}
 for(i=0;i<Stunnel_challeN;i++)
   challenge[i]= 0;
 for(i=0;i<2*Stunnel_challeN;i++) {
   c= toupper(cpt[i])-48;
   if(c>9)
     c-= 7;
   if(c<0||c>15)
     {ret= 0; goto ex;}
   if(i%2)
     challenge[i/2]|= c;
   else
     challenge[i/2]|= c<<4;
 }
 cpt+= i;

 /* read back channel */
 if(*cpt!=':')
   {ret= 0; goto ex;}
 Sregex_string(backchannel,cpt+1,0);

 ret= Stunnel_new(tunnel,user,idno,keydir,0);
 if(ret<=0) {
   if(ret==-2){
     sprintf(errmsg,"--- tunnel %ld : unknown user : %s",idno,user);
     ret= 0;
   }
   goto ex;
 }
 t= *tunnel;

 /* decipher challenge and make reply */
#ifndef Stunnel_no_fishinG
 Sblowfish_reset_chain(t->personal_fish,1);
 Sblowfish_transform(t->personal_fish,challenge,Stunnel_challeN,1);
#endif /* ! Stunnel_no_fishinG */
 for(i=0;i<Stunnel_challeN;i++)
   t->in_challenge[i]= challenge[i];
 Stunnel_make_reply(t,t->in_challenge,t->reply,0);
 ret= Stunnel_format(t,&reply,&reply_len,1,t->reply,Stunnel_challeN,1);
 if(ret<=0)
   goto ex;

 /* make session key and format it encrypted by personal_fish */
 Sblowfish_make_random(t->random_fish,&(t->random_count),
                       key,Stunnel_keyleN,0);
 ret= Sblowfish_new(&(t->session_fish),key,Stunnel_keyleN,0);
 if(ret<=0)
   goto ex;
 ret= Stunnel_format(t,&skey_reply,&skey_reply_len,2,key,Stunnel_keyleN,1);
 if(ret<=0)
   goto ex;
 ret= Sregex_concat_bin(&reply,&reply_len,skey_reply,skey_reply_len,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 if(ret<=0) {
   if(errmsg[0]==0) 
     if(*message_len<SfileadrL/2) {
       sprintf(errmsg,general_error,idno,message);
       Sfile_esc_filename(errmsg,sizeof(errmsg)-1,1|16);
     }
   if(t!=NULL)
     Stunnel_destroy(tunnel,0);
   Sregex_string(&reply,NULL,0);
 } else {
   if(*message!=NULL)
     Smem_freE((char *) *message);
   *message= reply;
   *message_len= reply_len;
 }
 Sregex_string(&gotstr,NULL,0);
 Sregex_string(&user,NULL,0);
 Sregex_string(&skey_reply,NULL,0);
 return(ret);
}


int Stunnel_handshake_accept(tunnel,message,message_len,endpt,flag)
struct StunneL *tunnel;
char *message;
int message_len;
char **endpt;  /* points to unconsumed message data for Stunnel_retrieve () */
int flag;
/*
return:
 <0 = error
  0 = authentication failed
  1 = handshake accepted
  2 = no decision yet, feed more data
*/
{
 int ret,i,type,untunneled_len= 0;
 char *cpt,*untunneled= NULL,expected_reply[Stunnel_challeN+1];

 cpt= message;
 /* decode message using personal_fish , accept any sequence number */
 ret= Stunnel_retrieve_token(tunnel,cpt,message_len-(cpt-message),&cpt,
                            &type,&untunneled,&untunneled_len,1|2);
 if(ret<=0)
   goto ex;

#ifdef Stunnel_no_formaT
 type= 1;
 untunneled_len= Stunnel_challeN;
 cpt= message+Stunnel_challeN;
#endif Stunnel_no_formaT

 /* compare with expected reply */
 if(untunneled_len!=Stunnel_challeN || type!=1)
   {ret=0; goto ex;}
 Stunnel_make_reply(tunnel,tunnel->challenge,expected_reply,0);
 for(i=0;i<Stunnel_challeN;i++)
   if(expected_reply[i]!=untunneled[i])
     break;
 if(i<Stunnel_challeN) {
   sprintf(tunnel->errmsg,
           "--- tunnel %ld : handshake reply does not match challenge.\n",
           tunnel->idno);
   ret=0;
   goto ex;
 }

 /* decode message and obtain session key */
 ret= Stunnel_retrieve_token(tunnel,cpt,message_len-(cpt-message),&cpt,
                             &type,&untunneled,&untunneled_len,1);
 if(ret<=0)
   goto ex;

#ifdef Stunnel_no_formaT
 type= 2;
 untunneled_len= Stunnel_keyleN;
 cpt= message+Stunnel_challeN+Stunnel_keyleN;
#endif /* Stunnel_no_formaT */

 if(untunneled_len!=Stunnel_keyleN || type!=2)
   {ret=0; goto ex;}

 /* create session_fish */
 ret= Sblowfish_new(&(tunnel->session_fish),untunneled,Stunnel_keyleN,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 *endpt= cpt;
 if(untunneled!=NULL)
   Smem_freE(untunneled);
 return(ret);
}


int Stunnel_reauthenticate(tunnel,message,messagelen,flag)
struct StunneL * tunnel;
char **message;
int *messagelen;
int flag;
{
 int ret;
 char msg[16];
 
 Sblowfish_make_random(tunnel->random_fish,&(tunnel->random_count),msg,16,0);
 ret= Stunnel_format(tunnel,message,messagelen,3,msg,16,0);
 return(ret);
}


int Stunnel_is_authenticated(tunnel,flag)
struct StunneL * tunnel;
int flag;
{
 if(tunnel==NULL)
   return(0);
 return(tunnel->authenticated);
}


int Stunnel_has_bytes(tunnel,flag)
struct StunneL * tunnel;
int flag;
{
 if(tunnel==NULL)
   return(0);
 if(tunnel->in_status==0)
   return(0);
 if(tunnel->in_status==1)
   return(tunnel->head_count);
 if(tunnel->in_status==2 || tunnel->in_status==3)
   return(tunnel->head_count+tunnel->data_count);
 return(0);
}


int Stunnel_make_userkey(tunnel,flag)
struct StunneL * tunnel;
int flag;
{
 char *adr= NULL,buffer[64];
 unsigned char key[Stunnel_keyleN],userfile[3*Stunnel_userleN+4];
 int ret,i,j;
 unsigned long crc[Stunnel_keyleN/4];

 strcpy(userfile,tunnel->user);
 if(strcmp(userfile,"internal_owner")==0)
   goto illegal_username;
 Sfile_esc_filename(userfile,sizeof(userfile)-1,1|4);
 if(strcmp(userfile,tunnel->user)!=0 || strchr(userfile,':')!=NULL) {
illegal_username:;
   sprintf(tunnel->errmsg,"illegal username '%s'",userfile);
   return(0);
 }
 if(Sregex_string(&adr,tunnel->keydir,0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&adr,"/",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&adr,userfile,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&adr,".tnl",1)<=0)
   {ret= -1; goto ex;}
 ret= Sfile_file_to_bin(adr,buffer,64,2);
 if(ret<=0) {
   sprintf(tunnel->errmsg,"cannot read 64 bytes from file  %s",adr);
   goto ex;
 }  

 /* compute key */
 for(i=0;i<Stunnel_keyleN/4;i++)
   crc[i]= 0;
 for(i=0;i<64;i++)
   Sregex_crc32(buffer+i,1,&(crc[i%(Stunnel_keyleN/4)]),1);
 for(i=0;i<Stunnel_keyleN/4;i++)
   for(j=0;j<4;j++)
    key[4*i+j]= (crc[i]>>(8*j))*0xff;

#ifndef Stunnel_no_fishinG
 ret= Sblowfish_transform(tunnel->internal_fish,key,Stunnel_keyleN,1|2);
 if(ret<=0)
   return(ret);
#endif /* ! Stunnel_no_fishinG */
 ret= Sfile_bin_to_file(adr,key,Stunnel_keyleN,1);
 if(ret<=0){
   sprintf(tunnel->errmsg,"cannot write key back to file  %s",adr);
   goto ex;
 }
 chmod(adr,S_IRUSR|S_IWUSR);
 ret= Stunnel_make_personal(tunnel,0);
 if(ret<=0)
   return(ret);
 ret= 1;
ex:;
 Sregex_string(&adr,NULL,0);
 return(ret);
}
